#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2026.01.06 06:00:00                  #
# ================================================== #

from typing import Optional, Union, Any, List

from PySide6.QtCore import QTimer
from PySide6.QtWidgets import QApplication, QFileDialog

from pygpt_net.utils import trans

class Batch:
    """
    Proxy for batch operations for a specific remote store provider used by Confirm router
    """

    def __init__(self, ctrl):
        self.ctrl = ctrl
        self.window = ctrl.window
        self.files_to_upload: List[str] = []

    # ---------- Starters (invoked by Confirm router) ----------

    def import_stores(self, force: bool = False):
        if not force:
            self.window.ui.dialogs.confirm(
                type=f'remote_store.import',
                id='',
                msg=trans('confirm.remote_store.import'),
            )
            return
        self.window.update_status("Importing stores...please wait...")
        self.ctrl._core_for(self._get_provider()).truncate()
        self._importer().import_vector_stores()

    def truncate_stores(self, force: bool = False):
        if not force:
            self.window.ui.dialogs.confirm(
                type=f'remote_store.truncate',
                id='',
                msg=trans('confirm.remote_store.truncate'),
            )
            return
        self.window.update_status("Removing stores...please wait...")
        QApplication.processEvents()
        self._importer().truncate_vector_stores()

    def refresh_stores(self, force: bool = False):
        if not force:
            self.window.ui.dialogs.confirm(
                type=f'remote_store.refresh',
                id='',
                msg=trans('confirm.remote_store.refresh'),
            )
            return
        self.window.update_status("Refreshing stores...please wait...")
        QApplication.processEvents()
        self._importer().refresh_vector_stores()

    def import_files_assistant_current(self, force: bool = False):
        """Import assistant files from API for current assistant"""
        id = self.window.core.config.get('assistant')
        if id is None or id == "":
            return
        if self.window.core.assistants.has(id):
            assistant = self.window.core.assistants.get_by_id(id)
            store_id = assistant.vector_store
            if store_id is None or store_id == "":
                self.window.ui.dialogs.alert(trans("dialog.remote_store.alert.assign"))
                return
            if not force:
                self.window.ui.dialogs.confirm(
                    type=f'remote_store.files.import.assistant.current',
                    id=store_id,
                    msg=trans('confirm.remote_store.import_files.store'),
                )
                return
            self.window.core.api.openai.store.importer.import_files(store_id) # for current assistant

    def import_files_assistant_all(self, force: bool = False):
        """Import assistant files from API for all assistants"""
        if not force:
            self.window.ui.dialogs.confirm(
                type=f'remote_store.files.import.assistant.all',
                id='',
                msg=trans('confirm.remote_store.import_files'),
            )
            return
        self.window.core.api.openai.store.importer.import_files()  # all

    def import_files(self, force: bool = False):
        if not force:
            self.window.ui.dialogs.confirm(
                type=f'remote_store.files.import.all',
                id='',
                msg=trans('confirm.remote_store.import_files'),
            )
            return
        self._importer().import_files()  # all

    def import_store_files(self, store_id: str, force: bool = False):
        if store_id is None:
            self.window.ui.dialogs.alert(trans("dialog.remote_store.alert.select"))
            return
        if not force:
            self.window.ui.dialogs.confirm(
                type=f'remote_store.files.import.store',
                id=store_id,
                msg=trans('confirm.remote_store.import_files.store'),
            )
            return
        self._importer().import_files(store_id)

    def truncate_files(self, force: bool = False):
        if not force:
            key = 'confirm.remote_store.files.truncate'
            self.window.ui.dialogs.confirm(
                type=f'remote_store.files.truncate',
                id='',
                msg=trans(key),
            )
            return
        self.window.update_status("Removing files...please wait...")
        QApplication.processEvents()
        self._importer().truncate_files()

    def truncate_store_files(self, store_id: Union[str, list], force: bool = False):
        if store_id is None:
            self.window.ui.dialogs.alert(trans("dialog.remote_store.alert.select"))
            return
        if not force:
            key = 'confirm.remote_store.files.truncate.store'
            self.window.ui.dialogs.confirm(
                type=f'remote_store.files.truncate.store',
                id=store_id,
                msg=trans(key),
            )
            return
        self.window.update_status("Removing files...please wait...")
        QApplication.processEvents()
        ids = store_id if isinstance(store_id, list) else [store_id]
        for sid in ids:
            self._importer().truncate_files(sid)

    def clear_stores(self, force: bool = False):
        if not force:
            self.window.ui.dialogs.confirm(
                type=f'remote_store.clear',
                id='',
                msg=trans('confirm.remote_store.clear'),
            )
            return
        self.window.update_status("Clearing stores...please wait...")
        QApplication.processEvents()
        self.ctrl._core_for(self._get_provider()).truncate()
        self.ctrl.update()
        self.ctrl.current = None
        self.ctrl.init()
        self.window.update_status("OK. All stores cleared.")
        if self._get_provider() == "openai":
            self.window.controller.assistant.files.update()

    def clear_files(self, force: bool = False):
        if not force:
            self.window.ui.dialogs.confirm(
                type=f'remote_store.files.clear.all',
                id='',
                msg=trans('confirm.assistant.files.clear'),
            )
            return
        self.window.update_status("Clearing documents...please wait...")
        QApplication.processEvents()
        self.ctrl._files_core_for(self._get_provider()).truncate_local()
        self.ctrl.update()
        self.window.update_status("OK. All documents cleared.")
        self.window.ui.dialogs.alert(trans("status.finished"))

    def clear_store_files(self, store_id: Optional[Union[str, list]] = None, force: bool = False):
        if store_id is None:
            self.window.ui.dialogs.alert(trans("dialog.remote_store.alert.select"))
            return
        if not force:
            self.window.ui.dialogs.confirm(
                type=f'remote_store.files.clear.store',
                id=store_id,
                msg=trans('confirm.assistant.files.clear'),
            )
            return
        self.window.update_status("Clearing store documents...please wait...")
        QApplication.processEvents()
        ids = store_id if isinstance(store_id, list) else [store_id]
        for sid in ids:
            self.ctrl._files_core_for(self._get_provider()).truncate_local(sid)
        self.ctrl.update()
        self.window.update_status("OK. Store documents cleared.")
        self.window.ui.dialogs.alert(trans("status.finished"))

    # ---------- Upload ----------

    def open_upload_files(self):
        if self.ctrl.current is None:
            self.window.ui.dialogs.alert(trans("dialog.remote_store.alert.select"))
            return
        options = QFileDialog.Options()
        files, _ = QFileDialog.getOpenFileNames(
            self.window,
            "Select file(s)...",
            "",
            "All Files (*)",
            options=options
        )
        if files:
            self.files_to_upload = files
        if self.files_to_upload:
            msg = "Are you sure you want to upload {} file(s)?".format(len(self.files_to_upload))
            self.window.ui.dialogs.confirm(
                type=f"remote_store.files.upload",
                id=0,
                msg=msg,
            )

    def open_upload_dir(self):
        if self.ctrl.current is None:
            self.window.ui.dialogs.alert(trans("dialog.remote_store.alert.select"))
            return
        options = QFileDialog.Options()
        directory = QFileDialog.getExistingDirectory(self.window, "Select directory...", options=options)
        if directory:
            self.files_to_upload = self.window.core.filesystem.get_files_from_dir(directory)
        if self.files_to_upload:
            msg = ("Are you sure you want to upload {} file(s) from directory {}?".
                   format(len(self.files_to_upload), directory))
            self.window.ui.dialogs.confirm(
                type=f"remote_store.files.upload",
                id=0,
                msg=msg,
            )

    def upload(self, force: bool = False):
        if self.ctrl.current is None:
            self.window.ui.dialogs.alert(trans("dialog.remote_store.alert.select"))
            return
        self.window.update_status("Uploading files...please wait...")
        QApplication.processEvents()
        self._importer().upload_files(self.ctrl.current, self.files_to_upload)
        self.files_to_upload = []

    def refresh_delayed(self, ms: int = 1000, all: bool = False):
        if all:
            self.window.update_status("Refreshing all stores...")
            QTimer.singleShot(ms, lambda: self.refresh_stores(force=True))
            return
        self.window.update_status("Refreshing status...")
        QTimer.singleShot(ms, lambda: self.ctrl.refresh_status())

    # ---------- Importer signal handlers (called by provider Importer) ----------

    def handle_refreshed_stores(self, num: int):
        self.ctrl.update()
        self.window.ui.dialogs.alert(trans("status.finished"))
        self.window.update_status("OK. All stores refreshed.")

    def handle_refreshed_stores_failed(self, error: Any):
        self.window.core.debug.log(error)
        self.ctrl.update()
        self.window.update_status("Error refreshing stores.")
        self.window.ui.dialogs.alert(error)

    def handle_imported_stores(self, num: int):
        self.ctrl.after_imported_stores(self._get_provider())
        self.ctrl.update()
        self.window.update_status("OK. Imported stores: " + str(num) + ".")

    def handle_imported_stores_failed(self, error: Any):
        self.window.core.debug.log(error)
        self.ctrl.update()
        self.window.update_status("Error importing stores.")
        self.window.ui.dialogs.alert(error)

    def handle_truncated_stores(self, num: int):
        # Clear local DB for provider
        core = self.ctrl._core_for(self._get_provider())
        core.truncate()
        self.ctrl.after_truncated_stores(self._get_provider())
        self.ctrl.update()
        self.ctrl.current = None
        self.ctrl.init()
        self.window.update_status("OK. Removed stores: " + str(num) + ".")
        self.window.ui.dialogs.alert(trans("status.finished"))

    def handle_truncated_stores_failed(self, error: Any):
        self.window.core.debug.log(error)
        self.ctrl.update()
        self.window.update_status("Error removing stores.")
        self.window.ui.dialogs.alert(error)

    def handle_imported_files(self, num: int):
        try:
            self.window.controller.assistant.files.update()
        except Exception:
            pass
        self.window.update_status("OK. Imported files: " + str(num) + ".")
        self.window.ui.dialogs.alert(trans("status.finished"))

    def handle_imported_files_failed(self, error: Any):
        self.window.core.debug.log(error)
        self.window.update_status("Error importing documents.")
        self.window.ui.dialogs.alert(error)

    def handle_truncated_files(self, store_id: Optional[str] = None, num: int = 0):
        self.window.update_status("OK. Truncated documents: " + str(num) + ".")
        try:
            if store_id :
                self.ctrl.refresh_by_store_id_provider(self._get_provider(), store_id)
            else:
                self.refresh_delayed(1200, all=True)
        except Exception as e:
            print(e)
        finally:
            self.window.ui.dialogs.alert(trans("status.finished"))

    def handle_truncated_files_failed(self, error: Any):
        self.window.core.debug.log(error)
        self.window.update_status("Error truncating documents.")
        self.window.ui.dialogs.alert(error)

    def handle_uploaded_files(self, num: int):
        try:
            self.window.controller.assistant.files.update()
        except Exception:
            pass
        self.window.update_status("OK. Uploaded files: " + str(num) + ".")
        self.window.ui.dialogs.alert("OK. Uploaded files: " + str(num) + ".")
        self.refresh_delayed(1500)

    def handle_uploaded_files_failed(self, error: Any):
        self.window.core.debug.log(error)
        self.refresh_delayed(1500)
        self.window.update_status("Error uploading files.")
        self.window.ui.dialogs.alert(error)

    # ---------- Internal ----------

    def _get_provider(self):
        return self.ctrl.provider_key

    def _importer(self):
        provider = self._get_provider()
        if (hasattr(self.window.core.api, provider)
                and hasattr(getattr(self.window.core.api, provider), 'store')
                and hasattr(getattr(self.window.core.api, provider).store, 'importer')):
            return getattr(self.window.core.api, provider).store.importer


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2026.01.07 23:00:00                  #
# ================================================== #

import asyncio
import base64
import json
from typing import Optional, Callable, Awaitable, Tuple, List, Any

from google.genai import types as gtypes  # for Schema/FunctionDeclaration/FunctionResponse compatibility

from pygpt_net.core.events import RealtimeEvent
from pygpt_net.core.types import MODE_AUDIO
from pygpt_net.item.ctx import CtxItem
from pygpt_net.core.text.utils import has_unclosed_code_tag

# shared
from pygpt_net.core.realtime.shared.loop import BackgroundLoop
from pygpt_net.core.realtime.shared.audio import to_pcm16_mono
from pygpt_net.core.realtime.shared.tools import build_function_responses_payload
from pygpt_net.core.realtime.shared.text import coalesce_text
from pygpt_net.core.realtime.shared.turn import TurnMode, apply_turn_mode_google
from pygpt_net.core.realtime.shared.session import set_ctx_rt_handle


class GoogleLiveClient:
    """
    Google Live client with server-side memory and smooth audio:

    - One persistent Live session; server keeps conversation context across turns.
    - User turns are sent via:
        * text: send_client_content(Content(...), turn_complete=True/False)
        * audio: ActivityStart -> send_realtime_input(audio=Blob...) -> ActivityEnd
          (manual turns; no auto VAD; no inline dicts — SDK serializes wire format)
    - Auto-turn mode (automatic VAD) is fully supported for continuous mic input:
        * push audio chunks via send_realtime_input(audio=...)
        * flush on demand via send_realtime_input(audio_stream_end=True)
        * receiver for one model turn is started automatically on first audio chunk.
    - Each turn has its own receive loop, ending on serverContent.turnComplete or toolCall.
    - Audio is jitter-buffered (~60ms) and de-duplicated (prefer response.data over inline_data).
    - Final transcript is coalesced; preserves hard line breaks only.
    - Tool calls, citations, images and usage are extracted and persisted to ctx to mirror OpenAI provider behavior.
    - Emits RealtimeEvent.RT_OUTPUT_AUDIO_COMMIT when the model starts responding after auto VAD or after an explicit flush,
      and RealtimeEvent.RT_OUTPUT_TURN_END after each turn.
    - Supports sending tool results back to the model (send_tool_results/send_tool_results_sync).
    """
    def __init__(
            self,
            window=None,
            debug: bool = False
    ):
        self.window = window
        self.debug = debug

        # Live session resources (owned by background loop)
        self._session = None
        self._session_cm = None

        # Background loop
        self._bg: BackgroundLoop = BackgroundLoop(name="Google-RT-Loop")

        # Flow control (per-session)
        self._send_lock: Optional[asyncio.Lock] = None
        self._response_done: Optional[asyncio.Event] = None
        self._response_active: bool = False
        self._turn_task: Optional[asyncio.Task] = None

        # Callbacks and context
        self._on_text: Optional[Callable[[str], Awaitable[None]]] = None
        self._on_audio: Optional[Callable[[bytes, str, Optional[int], Optional[int], bool], Awaitable[None]]] = None
        self._should_stop: Optional[Callable[[], bool]] = None
        self._ctx: Optional[CtxItem] = None
        self._last_opts = None

        # Per-turn text aggregation
        self._turn_text_parts: List[str] = []
        self._last_out_tr: str = ""  # last full output transcription (to compute deltas)

        # Audio I/O (rates)
        self._IN_RATE = 16000     # input (LINEAR16 mono)
        self._OUT_RATE = 24000    # output (model audio PCM16@24kHz)

        # Output audio jitter buffer
        self._audio_buf = bytearray()
        self._OUT_CHUNK_MS = 60
        self._OUT_BYTES_PER_MS = int(self._OUT_RATE * 2 / 1000)  # PCM16 mono (2 bytes/sample)
        self._saw_data_stream = False  # prefer response.data over inline_data to avoid duplicates

        # Per-turn extraction state
        self._rt_state: Optional[dict] = None

        # Last tool calls snapshot
        self._last_tool_calls: list[dict] = []

        # Live session resumption (current session handle)
        self._rt_session_id: Optional[str] = None  # string handle that can be used to resume a session

        # Cached tools signature to avoid redundant restarts
        self._cached_session_tools_sig: Optional[str] = None

        # Auto-turn state
        self._auto_audio_in_flight: bool = False  # True if auto-turn audio has been sent in current turn

    # -----------------------------
    # Public high-level entrypoints
    # -----------------------------

    async def run(
        self,
        ctx: CtxItem,
        opts,
        on_text: Callable[[str], Awaitable[None]],
        on_audio: Callable[[bytes, str, Optional[int], Optional[int], bool], Awaitable[None]],
        should_stop: Callable[[], bool] = lambda: False,
    ):
        """
        Run one turn: open session if needed, send prompt/audio, receive until turn complete.
        """
        self._ensure_background_loop()
        self._ctx = ctx

        # If a different resumable handle is provided, reset the session to resume there
        try:
            provided = getattr(opts, "rt_session_id", None)
            if isinstance(provided, str):
                provided = provided.strip()
            if self._session is not None and provided and provided != (self._rt_session_id or ""):
                await self._run_on_owner(self._reset_session_internal(ctx, opts, on_text, on_audio, should_stop))
        except Exception:
            pass

        if not self._session:
            await self._run_on_owner(self._open_session_internal(ctx, opts, on_text, on_audio, should_stop))

        await self._run_on_owner(self._send_turn_internal(
            getattr(opts, "prompt", None),
            getattr(opts, "audio_data", None),
            getattr(opts, "audio_format", None),
            getattr(opts, "audio_rate", None),
            wait_for_done=not bool(getattr(opts, "streaming", False)),
        ))

    async def open_session(
        self,
        ctx: CtxItem,
        opts,
        on_text: Callable[[str], Awaitable[None]],
        on_audio: Callable[[bytes, str, Optional[int], Optional[int], bool], Awaitable[None]],
        should_stop: Callable[[], bool] = lambda: False,
    ):
        """
        Open persistent Live session (if not already open).
        """
        self._ensure_background_loop()

        # If the session is already open but a different handle is requested, reset to resume.
        try:
            provided = getattr(opts, "rt_session_id", None)
            if isinstance(provided, str):
                provided = provided.strip()
            if self._session is not None and provided and provided != (self._rt_session_id or ""):
                await self._run_on_owner(self._reset_session_internal(ctx, opts, on_text, on_audio, should_stop))
                return
        except Exception:
            pass

        await self._run_on_owner(self._open_session_internal(ctx, opts, on_text, on_audio, should_stop))

    async def close_session(self):
        """Close persistent Live session (if open)."""
        if not self._bg.loop:
            return
        await self._run_on_owner(self._close_session_internal())

    async def reset_session(
        self,
        ctx: Optional[CtxItem] = None,
        opts=None,
        on_text: Optional[Callable[[str], Awaitable[None]]] = None,
        on_audio: Optional[Callable[[bytes, str, Optional[int], Optional[int], bool], Awaitable[None]]] = None,
        should_stop: Optional[Callable[[], bool]] = None,
    ):
        """
        Reset (close and reopen) persistent Live session with same or new params.
        """
        self._ensure_background_loop()
        await self._run_on_owner(self._reset_session_internal(ctx, opts, on_text, on_audio, should_stop))

    async def shutdown(self):
        """Shutdown background loop and close session."""
        if not self._bg.loop:
            return
        await self._run_on_owner(self._close_session_internal())

    async def shutdown_and_stop(self):
        """Shutdown background loop, close session and stop the loop thread."""
        await self.shutdown()
        self.stop_loop_sync()

    # -----------------------------
    # Synchronous convenience calls
    # -----------------------------

    def close_session_sync(self, timeout: float = 5.0):
        """Close persistent Live session (if open)."""
        if not self._bg.loop or not self._bg.loop.is_running():
            return
        self._bg.run_sync(self._close_session_internal(), timeout=timeout)

    def reset_session_sync(
        self,
        ctx: Optional[CtxItem] = None,
        opts=None,
        on_text: Optional[Callable[[str], Awaitable[None]]] = None,
        on_audio: Optional[Callable[[bytes, str, Optional[int], Optional[int], bool], Awaitable[None]]] = None,
        should_stop: Optional[Callable[[], bool]] = None,
        timeout: float = 10.0,
    ):
        """
        Reset (close and reopen) persistent Live session with same or new params.
        """
        self._ensure_background_loop()
        self._bg.run_sync(self._reset_session_internal(ctx, opts, on_text, on_audio, should_stop), timeout=timeout)

    def shutdown_sync(self, timeout: float = 5.0):
        """
        Shutdown background loop and close session (sync).
        """
        if not self._bg.loop or not self._bg.loop.is_running():
            return
        self._bg.run_sync(self._close_session_internal(), timeout=timeout)

    def stop_loop_sync(self, timeout: float = 2.0):
        """
        Stop background loop and join the thread.
        """
        self._bg.stop(timeout=timeout)

    # -----------------------------
    # Tools helpers
    # -----------------------------

    def _update_last_opts_tools(self, tools: Optional[list], remote_tools: Optional[list]) -> None:
        """
        Update self._last_opts with tools/remote_tools if those attributes exist.
        """
        lo = self._last_opts
        if not lo:
            return
        try:
            if tools is not None and hasattr(lo, "tools"):
                setattr(lo, "tools", tools)
        except Exception:
            pass
        try:
            if remote_tools is not None and hasattr(lo, "remote_tools"):
                setattr(lo, "remote_tools", remote_tools)
        except Exception:
            pass

    def _tools_signature(self, tools_list: list) -> str:
        """
        Build a stable signature string for the given tools list.
        """
        try:
            return json.dumps(tools_list or [], ensure_ascii=False, sort_keys=True, separators=(",", ":"))
        except Exception:
            return str(tools_list or [])

    # -----------------------------
    # Internal: background loop/dispatch
    # -----------------------------

    def _ensure_background_loop(self):
        """Ensure background event loop and thread are running."""
        self._bg.ensure()

    async def _run_on_owner(self, coro):
        """
        Run coroutine on the owner loop and await result.
        """
        return await self._bg.run(coro)

    # -----------------------------
    # Internal: session lifecycle
    # -----------------------------

    async def _open_session_internal(
        self,
        ctx: CtxItem,
        opts,
        on_text: Callable[[str], Awaitable[None]],
        on_audio: Callable[[bytes, str, Optional[int], Optional[int], bool], Awaitable[None]],
        should_stop: Callable[[], bool] = lambda: False,
    ):
        """
        Open persistent Live session (if not already open).
        """
        if self._session is not None:
            if self.debug:
                print("[google.open_session] already open")
            return

        core = self.window.core
        model_data = core.models.get(ctx.model) if ctx and getattr(ctx, "model", None) else None
        client = self.window.core.api.google.get_client(MODE_AUDIO, model_data if ctx else None)
        if not client:
            raise RuntimeError("Google GenAI client not configured")

        # Select Live-capable model
        model_id = getattr(opts, "model", None) or (ctx.model if ctx and getattr(ctx, "model", None) else "gemini-live-2.5-flash-preview")
        voice = getattr(opts, "voice", None) or self._preferred_voice()

        # Compose tools for session
        session_tools = self._sanitize_tools(getattr(opts, "tools", None), getattr(opts, "remote_tools", None))

        # Live config — manual activity boundaries (no auto VAD by default)
        live_cfg = {
            "response_modalities": ["AUDIO"],
            "speech_config": {"voice_config": {"prebuilt_voice_config": {"voice_name": voice}}},
            "output_audio_transcription": {},
            "realtime_input_config": {"automatic_activity_detection": {"disabled": True}},
        }
        if session_tools:
            live_cfg["tools"] = session_tools

        # Cache current tools signature
        self._cached_session_tools_sig = self._tools_signature(session_tools or [])

        sys_prompt = getattr(opts, "system_prompt", None)
        if sys_prompt:
            live_cfg["system_instruction"] = str(sys_prompt)

        # Save callbacks and ctx early so handle persistence can target the current context
        self._on_text = on_text
        self._on_audio = on_audio
        self._should_stop = should_stop or (lambda: False)
        self._ctx = ctx
        self._last_opts = opts

        # Session resumption: configure per docs; include handle when provided, otherwise None.
        try:
            ph = None
            provided_handle = getattr(opts, "rt_session_id", None)
            if isinstance(provided_handle, str):
                ph = provided_handle.strip() or None

            sr_cfg = gtypes.SessionResumptionConfig(handle=ph)
            live_cfg["session_resumption"] = sr_cfg

            if ph:
                self._persist_rt_handle(ph)
        except Exception:
            pass

        # Apply turn mode (auto/manual VAD)
        turn_mode = TurnMode.AUTO if bool(getattr(opts, "auto_turn", False)) else TurnMode.MANUAL
        apply_turn_mode_google(live_cfg, turn_mode)
        self._tune_google_vad(live_cfg, opts)

        # Control primitives
        self._response_done = asyncio.Event()
        self._send_lock = asyncio.Lock()
        self._turn_text_parts = []
        self._last_out_tr = ""
        self._last_tool_calls = []

        # Connect session
        self._session_cm = client.aio.live.connect(model=model_id, config=live_cfg)
        self._session = await self._session_cm.__aenter__()
        if self.debug:
            print("[google.open_session] live session connected")

    async def _close_session_internal(self):
        """Close persistent Live session (if open)."""
        if self._turn_task and not self._turn_task.done():
            try:
                await asyncio.wait_for(self._turn_task, timeout=2.0)
            except Exception:
                pass
        self._turn_task = None

        if self._session_cm:
            try:
                await self._session_cm.__aexit__(None, None, None)
            except Exception:
                pass
        self._session_cm = None
        self._session = None

        self._response_active = False
        self._response_done = None
        self._send_lock = None
        self._turn_text_parts = []
        self._last_out_tr = ""
        self._audio_buf.clear()
        self._saw_data_stream = False
        self._rt_state = None
        self._last_tool_calls = []

        # Clear in-memory handle as well to prevent unintended resumption
        self._rt_session_id = None

        # Clear cached tools signature
        self._cached_session_tools_sig = None

        # Auto-turn flags
        self._auto_audio_in_flight = False

        if self.debug:
            print("[google.close_session] closed")

    async def _reset_session_internal(
        self,
        ctx: Optional[CtxItem] = None,
        opts=None,
        on_text: Optional[Callable[[str], Awaitable[None]]] = None,
        on_audio: Optional[Callable[[bytes, str, Optional[int], Optional[int], bool], Awaitable[None]]] = None,
        should_stop: Optional[Callable[[], bool]] = None,
    ):
        """
        Reset (close and reopen) persistent Live session with same or new params.
        """
        ctx = ctx or self._ctx
        opts = opts or self._last_opts
        on_text = on_text or self._on_text
        on_audio = on_audio or self._on_audio
        should_stop = should_stop or self._should_stop or (lambda: False)

        if not (ctx and opts and on_text and on_audio):
            raise RuntimeError("reset_session requires previous or explicit ctx/opts/callbacks")

        await self._close_session_internal()
        await self._open_session_internal(ctx, opts, on_text, on_audio, should_stop)

    # -----------------------------
    # Internal: one "turn"
    # -----------------------------

    async def _send_turn_internal(
        self,
        prompt: Optional[str] = None,
        audio_data: Optional[bytes] = None,
        audio_format: Optional[str] = None,
        audio_rate: Optional[int] = None,
        wait_for_done: bool = True,
    ):
        """
        Send one turn: prompt and/or audio, receive until turn complete or tool call.
        """
        if not self._session:
            if self._ctx and self._last_opts:
                await self._open_session_internal(self._ctx, self._last_opts, self._on_text, self._on_audio, self._should_stop)
            else:
                raise RuntimeError("Session not open. Call open_session(...) first.")

        if self._send_lock is None:
            self._send_lock = asyncio.Lock()

        async with self._send_lock:
            if self._response_active and self._response_done:
                if self.debug:
                    print("[google.send_turn] waiting for previous response")
                await self._response_done.wait()

            # Reset per-turn collectors
            self._turn_text_parts = []
            self._last_out_tr = ""
            self._audio_buf.clear()
            self._saw_data_stream = False
            self._rt_reset_state()
            self._last_tool_calls = []
            self._auto_audio_in_flight = False

            # Normalize prompt/audio first to choose a stable path
            txt = str(prompt).strip() if prompt is not None else ""
            if txt == "...":
                txt = ""
            parts_t = [gtypes.Part(text=txt)] if txt else []

            pcm = b""
            rate = self._IN_RATE
            if audio_data:
                pcm, rate = to_pcm16_mono(audio_data, audio_format, audio_rate, target_rate=self._IN_RATE)

            has_text = bool(parts_t)
            has_audio = bool(pcm)

            # Branches
            if has_text and not has_audio:
                # TEXT-ONLY -> single Content, turn_complete=True
                await self._session.send_client_content(
                    turns=gtypes.Content(role="user", parts=parts_t),
                    turn_complete=True,
                )
                self._response_active = True
                if self._response_done is None:
                    self._response_done = asyncio.Event()
                else:
                    try:
                        self._response_done.clear()
                    except Exception:
                        self._response_done = asyncio.Event()
                self._turn_task = asyncio.create_task(self._recv_one_turn(), name="google-live-turn")

            elif has_audio and not has_text:
                # AUDIO-ONLY
                # If auto-turn is enabled, use auto-VAD path and flush with audio_stream_end.
                # Otherwise, use manual ActivityStart/End boundaries.
                use_auto = False
                try:
                    use_auto = bool(getattr(self._last_opts, "auto_turn", False))
                except Exception:
                    use_auto = False

                self._response_active = True
                if self._response_done is None:
                    self._response_done = asyncio.Event()
                else:
                    try:
                        self._response_done.clear()
                    except Exception:
                        self._response_done = asyncio.Event()

                # Start receiving before sending any audio
                self._turn_task = asyncio.create_task(self._recv_one_turn(), name="google-live-turn")

                if use_auto:
                    self._auto_audio_in_flight = True
                    # Auto-VAD: send a single audio blob and flush explicitly
                    try:
                        await self._session.send_realtime_input(
                            audio=gtypes.Blob(data=pcm, mime_type=f"audio/pcm;rate={int(rate)}")
                        )
                        await self._session.send_realtime_input(audio_stream_end=True)
                        self._emit_audio_commit_signal()  # fire once for explicit flush
                    except Exception as e:
                        if self.debug:
                            print(f"[google.audio:auto] send failed: {e!r}")
                        raise
                else:
                    # Manual activity: start -> audio -> end
                    await self._send_audio_realtime_manual(pcm, rate)

            elif has_text and has_audio:
                # TEXT + AUDIO in one user turn:
                # Respect the configured mode: in manual mode keep ActivityStart/End,
                # in auto-turn mode send text first and then treat audio as auto-VAD stream with explicit flush.
                use_auto = False
                try:
                    use_auto = bool(getattr(self._last_opts, "auto_turn", False))
                except Exception:
                    use_auto = False

                # 1) text opens the turn (turn_complete=False)
                await self._session.send_client_content(
                    turns=gtypes.Content(role="user", parts=parts_t),
                    turn_complete=False,
                )

                self._response_active = True
                if self._response_done is None:
                    self._response_done = asyncio.Event()
                else:
                    try:
                        self._response_done.clear()
                    except Exception:
                        self._response_done = asyncio.Event()

                # Start receiver, then send audio
                self._turn_task = asyncio.create_task(self._recv_one_turn(), name="google-live-turn")

                if use_auto:
                    self._auto_audio_in_flight = True
                    try:
                        await self._session.send_realtime_input(
                            audio=gtypes.Blob(data=pcm, mime_type=f"audio/pcm;rate={int(rate)}")
                        )
                        await self._session.send_realtime_input(audio_stream_end=True)
                        self._emit_audio_commit_signal()  # fire once for explicit flush
                    except Exception as e:
                        if self.debug:
                            print(f"[google.audio:auto+text] send failed: {e!r}")
                        raise
                else:
                    await self._send_audio_realtime_manual(pcm, rate)

            else:
                # nothing to send
                return

        if wait_for_done and self._turn_task:
            try:
                await self._turn_task
            except Exception:
                pass

    async def _send_audio_realtime_manual(self, pcm: bytes, rate: int):
        """
        Manual turn boundaries: ActivityStart -> audio chunks -> ActivityEnd.
        MIME must be audio/pcm;rate=RATE (no space).
        """
        if not pcm:
            return
        mime = f"audio/pcm;rate={int(rate)}"

        # Activity start
        try:
            await self._session.send_realtime_input(activity_start=gtypes.ActivityStart())
            if self.debug:
                print("[google.audio] activityStart")
        except Exception as e:
            if self.debug:
                print(f"[google.audio] activityStart failed: {e!r}")
            raise

        # ~100 ms chunks (for 16kHz -> 3200 bytes)
        bytes_per_ms = int(rate * 2 / 1000)  # 2 bytes per sample, mono
        chunk = max(bytes_per_ms * 100, 3200)
        for i in range(0, len(pcm), chunk):
            part = pcm[i:i + chunk]
            try:
                await self._session.send_realtime_input(
                    audio=gtypes.Blob(data=part, mime_type=mime)
                )
            except Exception as e:
                if self.debug:
                    print(f"[google.audio] payload send failed: {e!r}")
                raise

        # Activity end
        try:
            await self._session.send_realtime_input(activity_end=gtypes.ActivityEnd())
            if self.debug:
                print("[google.audio] activityEnd")
        except Exception as e:
            if self.debug:
                print(f"[google.audio] activityEnd failed: {e!r}")
            raise

    # -----------------------------
    # Internal: realtime audio input (auto-turn mode)
    # -----------------------------

    def rt_handle_audio_input_sync(self, event: RealtimeEvent, timeout: float = 0.5):
        """
        Synchronous entrypoint for continuous microphone input when auto-turn is enabled.
        Safe to call from any thread; schedules work on the background loop.
        """
        # Quick no-op if empty
        try:
            payload = getattr(event, "data", {}) or {}
            if isinstance(payload, dict) and "payload" in payload and isinstance(payload["payload"], dict):
                payload = payload["payload"]
            if not payload or not payload.get("data"):
                return
        except Exception:
            return

        self._ensure_background_loop()
        try:
            self._bg.run_sync(self._rt_handle_audio_input_internal(event), timeout=timeout)
        except Exception:
            # Never raise to caller from audio callback
            pass

    async def _rt_handle_audio_input_internal(self, event: RealtimeEvent):
        """
        Owner-loop implementation: push live audio to Gemini Live in auto-turn mode.
        """
        if not self._session:
            return
        try:
            if not bool(getattr(self._last_opts, "auto_turn", False)):
                # Only handle here when auto-turn is on; manual mode uses ActivityStart/End path.
                return
        except Exception:
            return

        # Extract normalized payload
        payload = getattr(event, "data", {}) or {}
        if isinstance(payload, dict) and "payload" in payload and isinstance(payload["payload"], dict):
            payload = payload["payload"]

        data: bytes = payload.get("data") or b""
        if not data:
            return
        mime = str(payload.get("mime") or "audio/pcm")
        rate = int(payload.get("rate") or 0) or self._IN_RATE
        channels = int(payload.get("channels") or 1)
        is_final = bool(payload.get("final", False))

        # Normalize to LINEAR16 mono @16kHz (Live API input native rate)
        fmt_hint = "pcm16" if mime.startswith("audio/pcm") else None
        try:
            pcm, norm_rate = to_pcm16_mono(data, fmt_hint, rate, target_rate=self._IN_RATE)
        except Exception:
            return

        # Ensure a receiver for this auto-turn is running before sending audio
        self._ensure_auto_receiver_started()

        # Mark that auto-turn audio has been sent in this turn
        self._auto_audio_in_flight = True

        # Send audio blob; Gemini Live handles VAD automatically in auto mode
        if self._send_lock is None:
            self._send_lock = asyncio.Lock()

        async with self._send_lock:
            try:
                await self._session.send_realtime_input(
                    audio=gtypes.Blob(data=pcm, mime_type=f"audio/pcm;rate={int(norm_rate)}")
                )
            except Exception:
                return

            # If stream end is flagged, flush server-side VAD buffer
            if is_final:
                try:
                    await self._session.send_realtime_input(audio_stream_end=True)
                    self._emit_audio_commit_signal()  # fire once for explicit flush
                except Exception:
                    pass

    def commit_audio_input_sync(self, timeout: float = 0.5):
        """
        Synchronous entrypoint to flush the input audio stream in auto-turn mode.
        This sends audio_stream_end to force the model to process current buffered audio.
        Safe to call from any thread.
        """
        self._ensure_background_loop()
        try:
            self._bg.run_sync(self._commit_audio_input_internal(), timeout=timeout)
        except Exception:
            # Never raise to caller
            pass

    async def _commit_audio_input_internal(self):
        """
        Owner-loop implementation: in auto-turn mode flush server-side VAD buffer.
        """
        if not self._session:
            return
        try:
            if not bool(getattr(self._last_opts, "auto_turn", False)):
                return
        except Exception:
            return

        # Ensure a receiver is running for this turn
        self._ensure_auto_receiver_started()

        if self._send_lock is None:
            self._send_lock = asyncio.Lock()
        async with self._send_lock:
            try:
                await self._session.send_realtime_input(audio_stream_end=True)
                self._emit_audio_commit_signal()  # fire once for explicit flush
            except Exception:
                pass

    def force_response_now_sync(self, timeout: float = 5.0):
        """
        Synchronously force the model to create a response from current input buffer (auto-turn).
        Internally sends audio_stream_end and ensures a receiver is running for the pending turn.
        """
        self._ensure_background_loop()
        try:
            self._bg.run_sync(self._force_response_now_internal(), timeout=timeout)
        except Exception:
            # Defensive: do not propagate errors to caller
            pass

    async def _force_response_now_internal(self):
        """
        Owner-loop: in auto-turn mode, flush current audio buffer and guarantee that a receive task
        for the current model turn is running. No-op in manual mode.
        """
        if not self._session:
            return
        try:
            if not bool(getattr(self._last_opts, "auto_turn", False)):
                return
        except Exception:
            return

        # Ensure a receiver is running for this turn
        self._ensure_auto_receiver_started()

        # Flush server-side buffer to force the model to respond
        if self._send_lock is None:
            self._send_lock = asyncio.Lock()
        async with self._send_lock:
            try:
                await self._session.send_realtime_input(audio_stream_end=True)
                self._emit_audio_commit_signal()  # fire once for explicit flush
            except Exception:
                pass

    async def _recv_one_turn(self):
        """Receive one turn until serverContent.turnComplete or toolCall."""
        if self.debug:
            print("[google._recv_one_turn] start")

        turn_finished = False

        try:
            async for response in self._session.receive():
                # 0) Session resumption updates (store last resumable handle)
                try:
                    sru = getattr(response, "session_resumption_update", None) or getattr(response, "sessionResumptionUpdate", None)
                    if sru:
                        # Prefer robustness: persist handle if present, regardless of 'resumable' flag inconsistencies
                        new_handle = self._extract_sru_handle(sru)
                        if isinstance(new_handle, str) and new_handle.strip():
                            self._persist_rt_handle(new_handle.strip())
                            if self.debug:
                                print(f"[google.live] session handle updated: {self._rt_session_id}")
                except Exception:
                    pass

                # 1) Usage (top-level)
                try:
                    um = getattr(response, "usage_metadata", None) or getattr(response, "usageMetadata", None)
                    if um:
                        self._rt_capture_google_usage(um)
                except Exception:
                    pass

                # 2) Preferred audio source: response.data (PCM16@24kHz)
                data = getattr(response, "data", None)
                if isinstance(data, (bytes, bytearray)):
                    # First output from model -> emit commit once (auto-turn only)
                    self._maybe_emit_auto_commit()
                    self._saw_data_stream = True
                    await self._audio_push(bytes(data), final=False)

                # 3) Server content
                sc = getattr(response, "server_content", None) or getattr(response, "serverContent", None)
                if sc:
                    # Any serverContent reaching here implies the model started processing;
                    # emit commit once if not yet emitted (auto-turn only).
                    self._maybe_emit_auto_commit()

                    # Output transcription (often cumulative)
                    out_tr = getattr(sc, "output_transcription", None) or getattr(sc, "outputTranscription", None)
                    if out_tr and getattr(out_tr, "text", None) and self._on_text:
                        full = str(out_tr.text)
                        delta = full[len(self._last_out_tr):] if full.startswith(self._last_out_tr) else full
                        self._last_out_tr = full
                        if delta.strip():
                            self._turn_text_parts.append(delta)
                            try:
                                await self._on_text(delta)
                            except Exception:
                                pass

                    # Optional: input transcription (handy in manual mode)
                    in_tr = getattr(sc, "input_transcription", None) or getattr(sc, "inputTranscription", None)
                    if in_tr and getattr(in_tr, "text", None) and self.debug:
                        print("[google.input_tr]", in_tr.text)

                    # Model turn parts
                    model_turn = getattr(sc, "model_turn", None) or getattr(sc, "modelTurn", None)
                    if model_turn:
                        parts = getattr(model_turn, "parts", None) or []
                        for p in parts:
                            # Function call parts
                            fc = getattr(p, "function_call", None) or (p.get("function_call") if isinstance(p, dict) else None)
                            if fc:
                                name = getattr(fc, "name", None) or (fc.get("name") if isinstance(fc, dict) else "")
                                args_obj = getattr(fc, "args", None) or (fc.get("args") if isinstance(fc, dict) else {})
                                args_dict = self._to_plain_dict(args_obj) or {}
                                self._rt_state["tool_calls"].append({
                                    "id": getattr(fc, "id", None) or "",
                                    "type": "function",
                                    "function": {
                                        "name": name or "",
                                        "arguments": json.dumps(args_dict, ensure_ascii=False),
                                    }
                                })
                                # self._rt_state["force_func_call"] = True
                                self._last_tool_calls = list(self._rt_state["tool_calls"])
                                turn_finished = True  # let the app run tools now

                            # Text part
                            txt = getattr(p, "text", None) or (p.get("text") if isinstance(p, dict) else None)
                            if txt and self._on_text:
                                s = str(txt)
                                self._turn_text_parts.append(s)
                                try:
                                    await self._on_text(s)
                                except Exception:
                                    pass

                            # Code execution parts
                            ex = getattr(p, "executable_code", None) or (p.get("executable_code") if isinstance(p, dict) else None)
                            if ex:
                                lang = (getattr(ex, "language", None) or "python").strip() or "python"
                                code_txt = (
                                    getattr(ex, "code", None) or
                                    getattr(ex, "program", None) or
                                    getattr(ex, "source", None) or
                                    ""
                                )
                                if not self._rt_state["is_code"]:
                                    hdr = f"\n\n**Code interpreter**\n```{lang.lower()}\n"
                                    self._turn_text_parts.append(hdr + str(code_txt))
                                    try:
                                        if self._on_text:
                                            await self._on_text(hdr + str(code_txt))
                                    except Exception:
                                        pass
                                    self._rt_state["is_code"] = True
                                else:
                                    self._turn_text_parts.append(str(code_txt))
                                    try:
                                        if self._on_text:
                                            await self._on_text(str(code_txt))
                                    except Exception:
                                        pass

                            cer = getattr(p, "code_execution_result", None) or (p.get("code_execution_result") if isinstance(p, dict) else None)
                            if cer and self._rt_state["is_code"]:
                                tail = "\n\n```\n-----------\n"
                                self._turn_text_parts.append(tail)
                                try:
                                    if self._on_text:
                                        await self._on_text(tail)
                                except Exception:
                                    pass
                                self._rt_state["is_code"] = False

                            # Inline images
                            inline = getattr(p, "inline_data", None) or (p.get("inline_data") if isinstance(p, dict) else None)
                            if inline:
                                mime = (getattr(inline, "mime_type", "") or (inline.get("mime_type") if isinstance(inline, dict) else "") or "").lower()
                                if mime.startswith("image/"):
                                    pdata = getattr(inline, "data", None) if not isinstance(inline, dict) else inline.get("data")
                                    try:
                                        img_bytes = None
                                        if isinstance(pdata, (bytes, bytearray)):
                                            img_bytes = bytes(pdata)
                                        elif isinstance(pdata, str):
                                            img_bytes = base64.b64decode(pdata)
                                        if img_bytes:
                                            save_path = self.window.core.image.gen_unique_path(self._ctx)
                                            with open(save_path, "wb") as f:
                                                f.write(img_bytes)
                                            self._rt_state["image_paths"].append(save_path)
                                            if not isinstance(self._ctx.images, list):
                                                self._ctx.images = []
                                            if save_path not in self._ctx.images:
                                                self._ctx.images.append(save_path)
                                    except Exception:
                                        pass

                        # Citations (grounding)
                        try:
                            self._collect_google_citations_from_server_content(sc)
                        except Exception:
                            pass

                    # Turn complete signal
                    try:
                        if bool(getattr(sc, "turn_complete", None) or getattr(sc, "turnComplete", None)):
                            turn_finished = True
                    except Exception:
                        pass

                # 4) Dedicated toolCall message
                tc = getattr(response, "tool_call", None) or getattr(response, "toolCall", None)
                if tc:
                    self._maybe_emit_auto_commit()  # ensure commit signaled before handing off to tools
                    fcs = getattr(tc, "function_calls", None) or getattr(tc, "functionCalls", None) or []
                    new_calls = []
                    for fc in fcs:
                        name = getattr(fc, "name", "") or ""
                        args_obj = getattr(fc, "args", {}) or {}
                        args_dict = self._to_plain_dict(args_obj) or {}
                        new_calls.append({
                            "id": getattr(fc, "id", "") or "",
                            "type": "function",
                            "function": {
                                "name": name,
                                "arguments": json.dumps(args_dict, ensure_ascii=False),
                            }
                        })
                    if new_calls:
                        seen = {(tc["function"]["name"], tc["function"]["arguments"]) for tc in self._rt_state["tool_calls"]}
                        for c in new_calls:
                            key = (c["function"]["name"], c["function"]["arguments"])
                            if key not in seen:
                                self._rt_state["tool_calls"].append(c)
                                seen.add(key)
                        # self._rt_state["force_func_call"] = True
                        self._last_tool_calls = list(self._rt_state["tool_calls"])
                    turn_finished = True

                if turn_finished:
                    break

            # Flush jitter buffer
            await self._audio_push(b"", final=True)

        except asyncio.CancelledError:
            try:
                await self._audio_push(b"", final=True)
            except Exception:
                pass
        except Exception as e:
            if self.debug:
                print(f"[google._recv_one_turn] exception: {e!r}")
            try:
                await self._audio_push(b"", final=True)
            except Exception:
                pass
        finally:
            # Persist textual output
            try:
                if self.window and self.window.core and self._ctx:
                    txt = coalesce_text(self._turn_text_parts)
                    if has_unclosed_code_tag(txt):
                        txt += "\n```"
                    if txt:
                        self._ctx.output = txt
                    # Tokens usage
                    up = (self._rt_state or {}).get("usage_payload") or {}
                    if up:
                        in_tok = up.get("in")
                        out_tok = up.get("out")
                        self._ctx.set_tokens(in_tok if in_tok is not None else (self._ctx.input_tokens or 0),
                                             out_tok if out_tok is not None else 0)
                        try:
                            if not isinstance(self._ctx.extra, dict):
                                self._ctx.extra = {}
                            self._ctx.extra["usage"] = {
                                "vendor": "google",
                                "input_tokens": in_tok,
                                "output_tokens": out_tok,
                                "reasoning_tokens": up.get("reasoning", 0),
                                "total_reported": up.get("total"),
                            }
                        except Exception:
                            pass

                    # Citations to ctx.urls
                    cites = (self._rt_state or {}).get("citations") or []
                    if cites:
                        if self._ctx.urls is None:
                            self._ctx.urls = []
                        for u in cites:
                            if u not in self._ctx.urls:
                                self._ctx.urls.append(u)

                    # Images to ctx.images
                    imgs = (self._rt_state or {}).get("image_paths") or []
                    if imgs:
                        if not isinstance(self._ctx.images, list):
                            self._ctx.images = []
                        for p in imgs:
                            if p not in self._ctx.images:
                                self._ctx.images.append(p)

                    # Unpack tool calls
                    tcs = (self._rt_state or {}).get("tool_calls") or []
                    if tcs:
                        for tc in tcs:
                            fn = tc.get("function") or {}
                            if isinstance(fn.get("arguments"), dict):
                                fn["arguments"] = json.dumps(fn["arguments"], ensure_ascii=False)
                        self._ctx.force_call = bool((self._rt_state or {}).get("force_func_call"))
                        self.window.core.debug.info("[google.live] Tool calls found, unpacking...")
                        self.window.core.command.unpack_tool_calls_chunks(self._ctx, tcs)

                    self.window.core.ctx.update_item(self._ctx)
            except Exception:
                pass

            # Mark done for waiters
            self._response_active = False
            if self._response_done:
                try:
                    self._response_done.set()
                except Exception:
                    pass

            # Emit end-of-turn event for audio pipeline symmetry with OpenAI
            try:
                if self._last_opts and hasattr(self._last_opts, "rt_signals"):
                    self._last_opts.rt_signals.response.emit(RealtimeEvent(RealtimeEvent.RT_OUTPUT_TURN_END, {
                        "ctx": self._ctx,
                    }))
            except Exception:
                pass

            # Reset per-turn state
            self._rt_state = None
            self._auto_audio_in_flight = False

            if self.debug:
                print("[google._recv_one_turn] done")

    # -----------------------------
    # Public: live tools update
    # -----------------------------

    async def update_session_tools(
        self,
        tools: Optional[list] = None,
        remote_tools: Optional[list] = None,
        force: bool = False,
    ):
        """
        Update session tools for Google Live.
        Since the Live API does not support mid-session tool config updates via SDK,
        this performs a safe session restart with best-effort resumption if the tools changed.
        If the session is not open, it only updates cached opts for the next open.
        """
        self._ensure_background_loop()
        return await self._run_on_owner(
            self._update_session_tools_internal(tools, remote_tools, force)
        )

    def update_session_tools_sync(
        self,
        tools: Optional[list] = None,
        remote_tools: Optional[list] = None,
        force: bool = False,
        timeout: float = 10.0,
    ):
        """Synchronous wrapper over update_session_tools()."""
        self._ensure_background_loop()
        return self._bg.run_sync(
            self._update_session_tools_internal(tools, remote_tools, force),
            timeout=timeout
        )

    async def _update_session_tools_internal(
        self,
        tools: Optional[list],
        remote_tools: Optional[list],
        force: bool,
    ):
        """
        Owner-loop implementation for tools update on Google Live.

        Strategy:
        - Sanitize and compute signature of the requested tools set.
        - If session is closed: update last opts and clear local cache.
        - If session is open and tools changed (or force=True):
            * Wait for any active response to finish.
            * Restart the Live session and request resumption using the last known handle.
        """
        # Prepare target tools (prefer explicit args, fallback to last opts)
        try:
            target_tools_raw = tools if tools is not None else getattr(self._last_opts, "tools", None)
        except Exception:
            target_tools_raw = None
        try:
            target_remote_raw = remote_tools if remote_tools is not None else getattr(self._last_opts, "remote_tools", None)
        except Exception:
            target_remote_raw = None

        session_tools = self._sanitize_tools(target_tools_raw, target_remote_raw)
        new_sig = self._tools_signature(session_tools or [])

        # If session is not open, just cache for next open
        if not self._session:
            self._update_last_opts_tools(tools, remote_tools)
            self._cached_session_tools_sig = None
            if self.debug:
                print("[google.update_session_tools] session not open; cached for next open")
            return

        # Skip if unchanged
        if not force and self._cached_session_tools_sig == new_sig:
            self._update_last_opts_tools(tools, remote_tools)
            if self.debug:
                print("[google.update_session_tools] no changes; skipping restart")
            return

        # Ensure previous response is finished
        if self._send_lock is None:
            self._send_lock = asyncio.Lock()
        async with self._send_lock:
            if self._response_active and self._response_done:
                if self.debug:
                    print("[google.update_session_tools] waiting for active response to finish")
                try:
                    await self._response_done.wait()
                except Exception:
                    pass

        # Persist new tools into last opts
        self._update_last_opts_tools(tools, remote_tools)

        # Try to resume the session state after restart if possible
        prev_handle = self._rt_session_id

        # Inject resumption handle into opts for the next open
        try:
            if self._last_opts is not None and prev_handle:
                setattr(self._last_opts, "rt_session_id", prev_handle)
        except Exception:
            pass

        if self.debug:
            print("[google.update_session_tools] restarting session to apply new tools")

        # Restart session with updated opts and best-effort resume
        await self._reset_session_internal(
            ctx=self._ctx,
            opts=self._last_opts,
            on_text=self._on_text,
            on_audio=self._on_audio,
            should_stop=self._should_stop,
        )

        # Cache new signature to suppress redundant restarts
        self._cached_session_tools_sig = new_sig

        if self.debug:
            print(f"[google.update_session_tools] session restarted; tools={len(session_tools)}")

    # -----------------------------
    # Public: send tool results back to the model
    # -----------------------------

    async def send_tool_results(
        self,
        results,
        continue_turn: bool = True,
        wait_for_done: bool = True,
    ):
        """
        Send tool results back to the Live session (FunctionResponse list).
        """
        self._ensure_background_loop()
        return await self._run_on_owner(
            self._send_tool_results_internal(results, continue_turn, wait_for_done)
        )

    def send_tool_results_sync(
        self,
        results,
        continue_turn: bool = True,
        wait_for_done: bool = True,
        timeout: float = 20.0,
    ):
        """
        Synchronous wrapper for send_tool_results().
        """
        self._ensure_background_loop()
        return self._bg.run_sync(
            self._send_tool_results_internal(results, continue_turn, wait_for_done),
            timeout=timeout
        )

    async def _send_tool_results_internal(
        self,
        results,
        continue_turn: bool,
        wait_for_done: bool,
    ):
        """
        Internal implementation of send_tool_results.
        """
        if not self._session:
            raise RuntimeError("Live session is not open")

        # Build neutral list and convert to gtypes.FunctionResponse[]
        try:
            neutral = build_function_responses_payload(results, self._last_tool_calls)
        except Exception as e:
            raise RuntimeError(f"Invalid tool results payload: {e}") from e

        if not neutral:
            return

        fn_responses = [
            gtypes.FunctionResponse(id=e.get("id") or "", name=e.get("name") or "", response=e.get("response") or {})
            for e in neutral
        ]

        if self._send_lock is None:
            self._send_lock = asyncio.Lock()
        async with self._send_lock:
            try:
                await self._session.send_tool_response(function_responses=fn_responses)
            except Exception as e:
                raise RuntimeError(f"send_tool_response failed: {e}") from e

        if continue_turn:
            self._turn_text_parts = []
            self._last_out_tr = ""
            self._audio_buf.clear()
            self._saw_data_stream = False
            self._rt_reset_state()

            self._response_active = True
            if self._response_done is None:
                self._response_done = asyncio.Event()
            else:
                try:
                    self._response_done.clear()
                except Exception:
                    self._response_done = asyncio.Event()

            self._turn_task = asyncio.create_task(self._recv_one_turn(), name="google-live-turn-followup")

            if wait_for_done:
                try:
                    await self._turn_task
                except Exception:
                    pass

    # -----------------------------
    # Helpers
    # -----------------------------

    def _preferred_voice(self) -> str:
        """
        Get preferred TTS voice from options or default.
        """
        try:
            v = self.window.core.plugins.get_option("audio_output", "google_genai_tts_voice")
            if v:
                mapping = {"kore": "Kore", "puck": "Puck", "charon": "Charon", "verse": "Verse", "legend": "Legend"}
                return mapping.get(str(v).strip().lower(), str(v))
        except Exception:
            pass
        return "Kore"

    async def _audio_push(self, data: bytes, final: bool = False):
        """
        Push audio data to the output callback in ~100 ms chunks.
        """
        if not self._on_audio:
            return
        if data:
            self._audio_buf.extend(data)
        threshold = self._OUT_BYTES_PER_MS * self._OUT_CHUNK_MS
        while len(self._audio_buf) >= threshold:
            chunk = self._audio_buf[:threshold]
            del self._audio_buf[:threshold]
            try:
                await self._on_audio(bytes(chunk), "audio/pcm", self._OUT_RATE, 1, False)
            except Exception:
                pass
        if final:
            if self._audio_buf:
                try:
                    await self._on_audio(bytes(self._audio_buf), "audio/pcm", self._OUT_RATE, 1, False)
                except Exception:
                    pass
                self._audio_buf.clear()
            try:
                await self._on_audio(b"", "audio/pcm", self._OUT_RATE, 1, True)
            except Exception:
                pass

    def _to_plain_dict(self, obj: Any) -> Any:
        """
        Convert various objects (pydantic, dataclass, etc) to plain dict recursively.
        """
        try:
            if hasattr(obj, "to_json_dict"):
                return obj.to_json_dict()
            if hasattr(obj, "model_dump"):
                return obj.model_dump()
            if hasattr(obj, "to_dict"):
                return obj.to_dict()
        except Exception:
            pass
        if isinstance(obj, dict):
            return {k: self._to_plain_dict(v) for k, v in obj.items()}
        if isinstance(obj, (list, tuple)):
            return [self._to_plain_dict(x) for x in obj]
        return obj

    def _rt_reset_state(self):
        """Reset per-turn realtime state."""
        self._rt_state = {
            "tool_calls": [],
            "citations": [],
            "files": [],
            "image_paths": [],
            "is_image": False,
            "is_code": False,
            "force_func_call": False,
            "usage_payload": {},
            "auto_commit_signaled": False,
        }

    def _rt_capture_google_usage(self, um_obj: Any):
        """
        Capture Google GenAI token usage from usage_metadata object.
        """
        if not um_obj or self._rt_state is None:
            return

        def as_int(v):
            try:
                if v is None:
                    return None
                return int(v)
            except Exception:
                try:
                    return int(float(v))
                except Exception:
                    return None

        prompt = (getattr(um_obj, "prompt_token_count", None)
                  or getattr(um_obj, "promptTokenCount", None)
                  or getattr(um_obj, "prompt_tokens", None)
                  or None)
        total = (getattr(um_obj, "total_token_count", None)
                 or getattr(um_obj, "totalTokenCount", None)
                 or getattr(um_obj, "total_tokens", None)
                 or None)
        candidates = (getattr(um_obj, "candidates_token_count", None)
                      or getattr(um_obj, "candidatesTokenCount", None)
                      or getattr(um_obj, "output_tokens", None)
                      or None)
        reasoning = (getattr(um_obj, "candidates_reasoning_token_count", None)
                     or getattr(um_obj, "candidatesReasoningTokenCount", None)
                     or getattr(um_obj, "reasoning_tokens", None)
                     or 0)
        p = as_int(prompt)
        t = as_int(total)
        c = as_int(candidates)
        r = as_int(reasoning) or 0
        out_total = max(0, (t or 0) - (p or 0)) if (t is not None and p is not None) else c
        self._rt_state["usage_payload"] = {"in": p, "out": out_total, "reasoning": r, "total": t}

    def _collect_google_citations_from_server_content(self, sc: Any):
        """
        Collect citations (URLs) from server_content grounding metadata and add to rt_state and ctx.urls.
        """
        if self._rt_state is None:
            return

        def add_url(url: Optional[str]):
            if not url or not isinstance(url, str):
                return
            u = url.strip()
            if not (u.startswith("http://") or u.startswith("https://")):
                return
            if u not in self._rt_state["citations"]:
                self._rt_state["citations"].append(u)
            try:
                if self._ctx:
                    if self._ctx.urls is None:
                        self._ctx.urls = []
                    if u not in self._ctx.urls:
                        self._ctx.urls.append(u)
            except Exception:
                pass

        gm = getattr(sc, "grounding_metadata", None) or getattr(sc, "groundingMetadata", None)
        if gm:
            atts = getattr(gm, "grounding_attributions", None) or getattr(gm, "groundingAttributions", None) or []
            try:
                for att in atts or []:
                    for path in (
                        "web.uri", "web.url", "source.web.uri", "source.web.url",
                        "source.uri", "source.url", "uri", "url",
                    ):
                        add_url(self._safe_get(att, path))
            except Exception:
                pass
            for path in (
                "search_entry_point.uri", "search_entry_point.url",
                "searchEntryPoint.uri", "searchEntryPoint.url",
                "search_entry_point.rendered_content_uri", "searchEntryPoint.rendered_content_uri",
            ):
                add_url(self._safe_get(gm, path))

        try:
            mt = getattr(sc, "model_turn", None) or getattr(sc, "modelTurn", None)
            parts = getattr(mt, "parts", None) or []
            for p in parts:
                pcm = self._safe_get(p, "citation_metadata") or self._safe_get(p, "citationMetadata")
                if pcm:
                    arr = (self._safe_get(pcm, "citation_sources")
                           or self._safe_get(pcm, "citationSources")
                           or self._safe_get(pcm, "citations") or []
                           )
                    for cit in arr or []:
                        for path in ("uri", "url", "source.uri", "source.url", "web.uri", "web.url"):
                            add_url(self._safe_get(cit, path))
                gpa = self._safe_get(p, "grounding_attributions") or self._safe_get(p, "groundingAttributions") or []
                for att in gpa or []:
                    for path in ("web.uri", "web.url", "source.web.uri", "source.web.url", "uri", "url"):
                        add_url(self._safe_get(att, path))
        except Exception:
            pass

    def _safe_get(self, obj, path: str) -> Any:
        """
        Safely get a nested attribute or dict key by dot-separated path.
        """
        cur = obj
        for seg in path.split("."):
            if cur is None:
                return None
            if isinstance(cur, dict):
                cur = cur.get(seg)
            else:
                if seg.isdigit() and isinstance(cur, (list, tuple)):
                    idx = int(seg)
                    if 0 <= idx < len(cur):
                        cur = cur[idx]
                    else:
                        return None
                else:
                    cur = getattr(cur, seg, None)
        return cur

    # -------- tools sanitizer for Live config (dict-only, robust) --------

    def _sanitize_tools(self, tools: Any, remote_tools: Optional[list] = None) -> list:
        """
        Normalize opts.tools into Live API config.tools (list of dicts).
        Supports gtypes.Tool, dict, or mixed list.
        """
        out: list = []
        sigset: set[str] = set()

        def add(entry: dict):
            try:
                sig = json.dumps(entry, ensure_ascii=False, sort_keys=True, separators=(",", ":"))
            except Exception:
                sig = str(entry)
            if sig not in sigset:
                out.append(entry)
                sigset.add(sig)

        def handle_tool_obj(t):
            # Convert gtypes.Tool -> dict
            fd_list = []
            fds = getattr(t, "function_declarations", None) or getattr(t, "functionDeclarations", None) or []
            for fd in fds or []:
                fd_dict = self._fd_to_dict(fd)
                if fd_dict:
                    fd_list.append(fd_dict)
            if fd_list:
                add({"function_declarations": fd_list})

            # built-ins
            if getattr(t, "code_execution", None) or getattr(t, "codeExecution", None):
                add({"code_execution": {}})
            if getattr(t, "google_search", None) or getattr(t, "googleSearch", None):
                add({"google_search": {}})
            if getattr(t, "url_context", None) or getattr(t, "urlContext", None):
                add({"url_context": {}})

        def handle_tool_dict(d: dict):
            fds = d.get("function_declarations") or d.get("functionDeclarations")
            if isinstance(fds, list) and fds:
                fd_list = []
                for fd in fds:
                    fd_dict = self._fd_to_dict(fd)
                    if fd_dict:
                        fd_list.append(fd_dict)
                if fd_list:
                    add({"function_declarations": fd_list})

            if (d.get("type") or "").lower() == "function":
                fn = d.get("function") if isinstance(d.get("function"), dict) else d
                name = fn.get("name")
                if name:
                    fd = {"name": str(name)}
                    if fn.get("description"):
                        fd["description"] = fn["description"]
                    params = fn.get("parameters")
                    fd["parameters"] = self._schema_to_plain(params if params is not None else {"type": "OBJECT"})
                    add({"function_declarations": [fd]})

            for k in ("google_search", "code_execution", "url_context"):
                if k in d and isinstance(d[k], dict):
                    add({k: dict(d[k])})
                elif k in d and d[k] is True:
                    add({k: {}})

        if isinstance(tools, (list, tuple)):
            for t in tools:
                if t is None:
                    continue
                if t.__class__.__name__ == "Tool" or isinstance(t, getattr(gtypes, "Tool", ())):
                    handle_tool_obj(t)
                elif isinstance(t, dict):
                    handle_tool_dict(t)

        if isinstance(remote_tools, (list, tuple)):
            for t in remote_tools:
                if isinstance(t, dict):
                    handle_tool_dict(t)
                elif t.__class__.__name__ == "Tool" or isinstance(t, getattr(gtypes, "Tool", ())):
                    handle_tool_obj(t)

        return out

    def _fd_to_dict(self, fd: Any) -> Optional[dict]:
        """
        Convert FunctionDeclaration (gtypes or dict) to plain dict with normalized schema.
        """
        if fd.__class__.__name__ == "FunctionDeclaration" or isinstance(fd, getattr(gtypes, "FunctionDeclaration", ())):
            name = getattr(fd, "name", None)
            if not name:
                return None
            out = {"name": str(name)}
            desc = getattr(fd, "description", None)
            if desc:
                out["description"] = desc
            params = getattr(fd, "parameters", None)
            out["parameters"] = self._schema_to_plain(params if params is not None else {"type": "OBJECT"})
            return out

        if isinstance(fd, dict):
            name = fd.get("name")
            if not name:
                return None
            out = {"name": str(name)}
            if fd.get("description"):
                out["description"] = fd["description"]
            params = fd.get("parameters")
            out["parameters"] = self._schema_to_plain(params if params is not None else {"type": "OBJECT"})
            return out

        return None

    def _schema_to_plain(self, sc: Any) -> dict:
        """
        Convert gtypes.Schema or dict to a plain dict acceptable by Live API.
        """
        allowed = {"OBJECT", "ARRAY", "STRING", "NUMBER", "INTEGER", "BOOLEAN"}
        alias = {"INT": "INTEGER", "BOOL": "BOOLEAN", "FLOAT": "NUMBER", "DOUBLE": "NUMBER"}

        def norm_type(val) -> str:
            n = getattr(val, "name", None)
            if isinstance(n, str) and n:
                s = n
            else:
                s = str(val or "")
                if "." in s:
                    s = s.split(".")[-1]
            s = alias.get(s.upper(), s.upper())
            return s if s in allowed else "OBJECT"

        if sc is not None and (sc.__class__.__name__ == "Schema" or isinstance(sc, getattr(gtypes, "Schema", ()))):
            d: dict = {}
            t = getattr(sc, "type", None)
            d["type"] = norm_type(t) if t is not None else "OBJECT"

            desc = getattr(sc, "description", None)
            if desc:
                d["description"] = desc
            fmt = getattr(sc, "format", None)
            if fmt:
                d["format"] = fmt
            enum = getattr(sc, "enum", None)
            if isinstance(enum, list) and enum and d["type"] == "STRING":
                d["enum"] = enum
            req = getattr(sc, "required", None)
            if isinstance(req, list) and req:
                d["required"] = [x for x in req if isinstance(x, str)]

            props = getattr(sc, "properties", None)
            if isinstance(props, dict) and props:
                d["properties"] = {k: self._schema_to_plain(v) for k, v in props.items()}

            items = getattr(sc, "items", None)
            if items:
                d["items"] = self._schema_to_plain(items)

            return d

        if isinstance(sc, dict):
            d = dict(sc)
            d["type"] = norm_type(d.get("type"))
            if isinstance(d.get("properties"), dict):
                d["properties"] = {k: self._schema_to_plain(v) for k, v in d["properties"].items()}
            if isinstance(d.get("items"), dict):
                d["items"] = self._schema_to_plain(d["items"])
            if "enum" in d and d.get("type") != "STRING":
                d.pop("enum", None)
            return d

        return {"type": "OBJECT"}

    def _tune_google_vad(self, live_cfg: dict, opts) -> None:
        """
        Increase end-of-speech hold for automatic VAD in Gemini Live.
        """
        try:
            ric = live_cfg.setdefault("realtime_input_config", {})
            aad = ric.setdefault("automatic_activity_detection", {})
            if aad.get("disabled") is True:
                return  # manual mode, VAD disabled

            # Resolve target silence (default 2000 ms)
            target_ms = getattr(opts, "vad_end_silence_ms", None)
            if not isinstance(target_ms, (int, float)) or target_ms <= 0:
                base = int(aad.get("silence_duration_ms") or 100)
                target_ms = max(base, 2000)

            aad["silence_duration_ms"] = int(target_ms)

            # Optional: make end-of-speech less aggressive
            try:
                aad["end_of_speech_sensitivity"] = gtypes.EndSensitivity.END_SENSITIVITY_LOW
            except Exception:
                aad["end_of_speech_sensitivity"] = "END_SENSITIVITY_LOW"

            # Optional: leading padding before detected speech
            prefix_ms = getattr(opts, "vad_prefix_padding_ms", None)
            if isinstance(prefix_ms, (int, float)) and prefix_ms >= 0:
                aad["prefix_padding_ms"] = int(prefix_ms)
        except Exception:
            pass

    def set_debug(self, enabled: bool):
        """
        Enable or disable debug logging.

        :param enabled: True to enable debug logging, False to disable.
        """
        self.debug = bool(enabled)

    def is_session(self) -> bool:
        """Check if the WS session is currently open."""
        return self._session is not None

    def is_session_active(self) -> bool:
        """Check if the WS session is currently open."""
        return self._session is not None

    def update_ctx(self, ctx: CtxItem):
        """Update the current CtxItem (for session handle persistence)."""
        self._ctx = ctx

    def get_current_rt_session_id(self) -> Optional[str]:
        """
        Return the current resumable session handle if known.
        """
        return self._rt_session_id

    # -----------------------------
    # Internal: auto-turn receiver bootstrap
    # -----------------------------

    def _ensure_auto_receiver_started(self):
        """
        Start a receiver task for one model turn in auto-turn mode if not already active.
        This guarantees we do not miss server responses when sending live audio chunks.
        """
        # Only in auto-turn mode and with an open session
        if not self._session:
            return
        try:
            if not bool(getattr(self._last_opts, "auto_turn", False)):
                return
        except Exception:
            return

        # If a previous task exists but is done, clear the ref
        if self._turn_task and self._turn_task.done():
            self._turn_task = None

        if not self._response_active:
            # Reset per-turn collectors
            self._turn_text_parts = []
            self._last_out_tr = ""
            self._audio_buf.clear()
            self._saw_data_stream = False
            self._rt_reset_state()

            self._response_active = True
            if self._response_done is None:
                self._response_done = asyncio.Event()
            else:
                try:
                    self._response_done.clear()
                except Exception:
                    self._response_done = asyncio.Event()

            self._turn_task = asyncio.create_task(self._recv_one_turn(), name="google-live-auto-turn")

    def update_session_autoturn_sync(
            self,
            enabled: bool,
            silence_ms: Optional[int] = None,
            prefix_ms: Optional[int] = None,
            timeout: float = 10.0,
    ):
        """
        Synchronous helper: enable/disable auto-turn (VAD) for Google Live
        and optionally override silence/prefix (milliseconds).
        Note: Live API doesn't support mid-session VAD reconfigure; we restart
        the session safely if it is open.
        """
        self._ensure_background_loop()
        return self._bg.run_sync(
            self._update_session_autoturn_internal(enabled, silence_ms, prefix_ms),
            timeout=timeout
        )

    async def _update_session_autoturn_internal(
            self,
            enabled: bool,
            silence_ms: Optional[int] = None,
            prefix_ms: Optional[int] = None,
    ):
        """
        Owner-loop: toggle auto-turn (automatic_activity_detection) and optionally
        set silence_duration_ms / prefix_padding_ms. If the session is open,
        perform a safe restart to apply new config. If closed, cache in opts.
        """

        # Helper to update cached opts
        def _apply_to_opts():
            if not self._last_opts:
                return
            try:
                setattr(self._last_opts, "auto_turn", bool(enabled))
            except Exception:
                pass
            try:
                if silence_ms is not None:
                    setattr(self._last_opts, "vad_end_silence_ms", int(silence_ms))
            except Exception:
                pass
            try:
                if prefix_ms is not None:
                    setattr(self._last_opts, "vad_prefix_padding_ms", int(prefix_ms))
            except Exception:
                pass

        # If session not open -> just cache and exit
        if not self._session:
            _apply_to_opts()
            if self.debug:
                print("[google.update_session_autoturn] session not open; cached for next open")
            return

        # Compute whether anything changes to avoid unnecessary restart
        cur_enabled = False
        try:
            cur_enabled = bool(getattr(self._last_opts, "auto_turn", False))
        except Exception:
            pass
        cur_sil = getattr(self._last_opts, "vad_end_silence_ms", None)
        cur_pre = getattr(self._last_opts, "vad_prefix_padding_ms", None)

        change = (cur_enabled != bool(enabled))
        if silence_ms is not None and int(silence_ms) != (int(cur_sil) if isinstance(cur_sil, (int, float)) else None):
            change = True
        if prefix_ms is not None and int(prefix_ms) != (int(cur_pre) if isinstance(cur_pre, (int, float)) else None):
            change = True

        if not change:
            # Nothing to do; still persist values to opts for consistency
            _apply_to_opts()
            if self.debug:
                print("[google.update_session_autoturn] no changes; skipping restart")
            return

        # Wait for any active response to finish before restart
        if self._send_lock is None:
            self._send_lock = asyncio.Lock()
        async with self._send_lock:
            if self._response_active and self._response_done:
                if self.debug:
                    print("[google.update_session_autoturn] waiting for active response to finish")
                try:
                    await self._response_done.wait()
                except Exception:
                    pass

        # Update cached opts with requested values
        _apply_to_opts()

        # Try to resume after restart using the last known handle (best-effort)
        prev_handle = self._rt_session_id
        try:
            if self._last_opts is not None and prev_handle:
                setattr(self._last_opts, "rt_session_id", prev_handle)
        except Exception:
            pass

        if self.debug:
            eff_sil = silence_ms if silence_ms is not None else cur_sil
            eff_pre = prefix_ms if prefix_ms is not None else cur_pre
            print(f"[google.update_session_autoturn] restarting session; auto_turn={enabled}, "
                  f"silence_ms={eff_sil}, prefix_ms={eff_pre}")

        # Restart session with updated config
        await self._reset_session_internal(
            ctx=self._ctx,
            opts=self._last_opts,
            on_text=self._on_text,
            on_audio=self._on_audio,
            should_stop=self._should_stop,
        )

        if self.debug:
            print("[google.update_session_autoturn] session restarted with new VAD settings")

    # -----------------------------
    # Internal: commit event helpers
    # -----------------------------

    def _emit_audio_commit_signal(self):
        """
        Emit RT_OUTPUT_AUDIO_COMMIT once per turn in auto-turn mode.
        """
        if self._rt_state is None:
            self._rt_reset_state()
        if self._rt_state.get("auto_commit_signaled"):
            return
        try:
            if not bool(getattr(self._last_opts, "auto_turn", False)):
                return
        except Exception:
            return
        # Limit to audio turns: only when we actually sent auto-turn audio this turn
        if not self._auto_audio_in_flight:
            return
        try:
            if self._last_opts and hasattr(self._last_opts, "rt_signals"):
                self._last_opts.rt_signals.response.emit(
                    RealtimeEvent(RealtimeEvent.RT_OUTPUT_AUDIO_COMMIT, {"ctx": self._ctx})
                )
            self._rt_state["auto_commit_signaled"] = True
        except Exception:
            pass

    def _maybe_emit_auto_commit(self):
        """
        Emit RT_OUTPUT_AUDIO_COMMIT on first sign of model output in auto-turn mode.
        """
        self._emit_audio_commit_signal()

    # -----------------------------
    # Internal: session handle helpers
    # -----------------------------

    def _persist_rt_handle(self, handle: str) -> None:
        """
        Persist current session handle in-memory, to ctx.extra and into last opts for future restarts.
        """
        try:
            self._rt_session_id = handle
            set_ctx_rt_handle(self._ctx, handle, self.window)
        except Exception:
            pass
        try:
            if self._last_opts is not None:
                setattr(self._last_opts, "rt_session_id", handle)
        except Exception:
            pass

    def _extract_sru_handle(self, sru: Any) -> Optional[str]:
        """
        Extract handle from SessionResumptionUpdate (supports snake_case and camelCase, and token alias).
        """
        # Objects (attrs)
        for attr in ("new_handle", "newHandle", "token"):
            try:
                v = getattr(sru, attr, None)
                if isinstance(v, str) and v.strip():
                    return v.strip()
            except Exception:
                pass
        # Dicts
        if isinstance(sru, dict):
            for k in ("new_handle", "newHandle", "token"):
                v = sru.get(k)
                if isinstance(v, str) and v.strip():
                    return v.strip()
        return None


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.15 23:00:00                  #
# ================================================== #

from pygpt_net.utils import trans


class Plugins:
    def __init__(self, window=None):
        """
        Plugins locale controller

        :param window: Window instance
        """
        self.window = window

    def apply(self):
        """Apply locale to plugins"""

        # plugins: info
        self.window.controller.plugins.update_info()

        plugins_dict = self.window.core.plugins.plugins
        plugin_ids = tuple(plugins_dict.keys())
        win = self.window
        ui = win.ui
        ui_nodes = ui.nodes
        ui_tabs = ui.tabs
        settings_tab = ui_tabs['plugin.settings']
        ui_menu_plugins = ui.menu.get('plugins', {})
        ui_config = ui.config
        ctrl_plugins = win.controller.plugins

        for plugin_id in plugin_ids:
            plugin = plugins_dict[plugin_id]
            if not plugin.use_locale:
                continue
            domain = f'plugin.{plugin_id}'
            trans('', True, domain)

        for plugin_id in plugin_ids:
            plugin = plugins_dict[plugin_id]
            if not plugin.use_locale:
                continue
            domain = f'plugin.{plugin_id}'

            name_txt = trans('plugin.name', False, domain)

            plugin_settings_desc_key = f'plugin.settings.{plugin_id}.desc'
            if plugin_settings_desc_key in ui_nodes:
                desc_txt = trans('plugin.description', False, domain)
                ui_nodes[plugin_settings_desc_key].setText(desc_txt)

            tab_idx = ctrl_plugins.get_tab_idx(plugin_id)
            if tab_idx is not None:
                settings_tab.setTabText(tab_idx, name_txt)

            if plugin_id in ui_menu_plugins:
                ui_menu_plugins[plugin_id].setText(name_txt)

            options = plugin.setup()
            if not options:
                continue

            cfg_domain = ui_config.get(domain)
            for option_id, option in options.items():
                label_key = f'plugin.{plugin_id}.{option_id}.label'
                desc_key = f'plugin.{plugin_id}.{option_id}.desc'

                is_bool = option.get('type') == 'bool'
                need_label = (label_key in ui_nodes) or (is_bool and cfg_domain and option_id in cfg_domain)
                need_desc = desc_key in ui_nodes

                label_str = None
                if need_label:
                    label_str = trans(f'{option_id}.label', False, domain)

                if need_desc:
                    desc_str = trans(f'{option_id}.description', False, domain)
                    tooltip_str = trans(f'{option_id}.tooltip', False, domain)
                    if tooltip_str == f'{option_id}.tooltip':
                        tooltip_str = desc_str
                    ui_nodes[desc_key].setText(desc_str)
                    ui_nodes[desc_key].setToolTip(tooltip_str)

                if label_key in ui_nodes and label_str is not None:
                    ui_nodes[label_key].setText(label_str)

                if is_bool and cfg_domain and option_id in cfg_domain and label_str is not None:
                    widget = cfg_domain[option_id]
                    if hasattr(widget, 'setText'):
                        widget.setText(label_str)
                    box = getattr(widget, 'box', None)
                    if box and hasattr(box, 'setText'):
                        box.setText(label_str)

        idx = settings_tab.currentIndex()
        win.plugin_settings.update_list('plugin.list', plugins_dict)
        ctrl_plugins.set_by_tab(idx)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2026.01.04 19:00:00                  #
# ================================================== #

import base64
import datetime
import os
from typing import Optional, Dict, Any, List, Iterable

import requests
from PySide6.QtCore import QObject, Signal, QRunnable, Slot

from pygpt_net.core.events import KernelEvent
from pygpt_net.core.bridge.context import BridgeContext
from pygpt_net.item.ctx import CtxItem
from pygpt_net.utils import trans


class Image:
    def __init__(self, window=None):
        self.window = window
        self.worker = None

    def generate(
            self,
            context: BridgeContext,
            extra: Optional[Dict[str, Any]] = None,
            sync: bool = True
    ) -> bool:
        """
        Generate image(s) via xAI SDK image API.
        Model: grok-2-image (or -1212 variants).

        :param context: BridgeContext with prompt, model, ctx
        :param extra: Extra parameters (num: int, inline: bool, etc.)
        :param sync: Run synchronously (blocking) if True
        :return: True if started
        """
        extra = extra or {}
        ctx = context.ctx or CtxItem()
        model = context.model
        prompt = context.prompt
        num = int(extra.get("num", 1))
        inline = bool(extra.get("inline", False))
        extra_prompt = extra.get("extra_prompt", "")

        # Optional prompt enhancement model (same as in your Google path)
        prompt_model = self.window.core.models.from_defaults()
        tmp = self.window.core.config.get('img_prompt_model')
        if self.window.core.models.has(tmp):
            prompt_model = self.window.core.models.get(tmp)

        worker = ImageWorker()
        worker.window = self.window
        worker.ctx = ctx
        worker.model = (model.id or "grok-2-image")
        worker.input_prompt = prompt
        worker.model_prompt = prompt_model
        worker.system_prompt = self.window.core.prompt.get('img')
        worker.raw = self.window.core.config.get('img_raw')
        worker.num = num
        worker.inline = inline
        worker.extra_prompt = extra_prompt

        self.worker = worker
        self.worker.signals.finished.connect(self.window.core.image.handle_finished)
        self.worker.signals.finished_inline.connect(self.window.core.image.handle_finished_inline)
        self.worker.signals.status.connect(self.window.core.image.handle_status)
        self.worker.signals.error.connect(self.window.core.image.handle_error)

        if sync or not self.window.controller.kernel.async_allowed(ctx):
            self.worker.run()
            return True

        self.window.dispatch(KernelEvent(KernelEvent.STATE_BUSY, {"id": "img"}))
        self.window.threadpool.start(self.worker)
        return True


class ImageSignals(QObject):
    finished = Signal(object, list, str)         # ctx, paths, prompt
    finished_inline = Signal(object, list, str)  # ctx, paths, prompt
    status = Signal(object)                      # message
    error = Signal(object)                       # exception


class ImageWorker(QRunnable):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.signals = ImageSignals()
        self.window = None
        self.ctx: Optional[CtxItem] = None

        # params
        self.model = "grok-2-image"
        self.model_prompt = None
        self.input_prompt = ""
        self.system_prompt = ""
        self.inline = False
        self.extra_prompt: Optional[str] = None
        self.raw = False
        self.num = 1

        # SDK image_format:
        # - "base64": returns raw image bytes in SDK response (preferred for local saving)
        # - "url": returns URL on xAI managed storage (fallback: we download)
        self.image_format = "base64"

    @Slot()
    def run(self):
        try:
            # optional prompt enhancement
            if not self.raw and not self.inline and self.input_prompt:
                try:
                    self.signals.status.emit(trans('img.status.prompt.wait'))
                    bridge_context = BridgeContext(
                        prompt=self.input_prompt,
                        system_prompt=self.system_prompt,
                        model=self.model_prompt,
                        max_tokens=200,
                        temperature=1.0,
                    )
                    ev = KernelEvent(KernelEvent.CALL, {'context': bridge_context, 'extra': {}})
                    self.window.dispatch(ev)
                    resp = ev.data.get('response')
                    if resp:
                        self.input_prompt = resp
                except Exception as e:
                    self.signals.error.emit(e)
                    self.signals.status.emit(trans('img.status.prompt.error') + ": " + str(e))

            # Negative prompt fallback: append as textual instruction (xAI has no native field for it)
            if self.extra_prompt and str(self.extra_prompt).strip():
                try:
                    self.input_prompt = self._merge_negative_prompt(self.input_prompt or "", self.extra_prompt)
                except Exception:
                    pass

            self.signals.status.emit(trans('img.status.generating') + f": {self.input_prompt}...")

            # use xAI SDK client
            client = self.window.core.api.xai.get_client()

            # enforce n range [1..10] as per docs
            n = max(1, min(int(self.num or 1), 10))

            images_bytes: List[bytes] = []
            if n == 1:
                # single image
                resp = client.image.sample(
                    model=self.model or "grok-2-image",
                    prompt=self.input_prompt or "",
                    image_format=("base64" if self.image_format == "base64" else "url"),
                )
                images_bytes = self._extract_bytes_from_single(resp)
            else:
                # batch images
                resp_iter = client.image.sample_batch(
                    model=self.model or "grok-2-image",
                    prompt=self.input_prompt or "",
                    n=n,
                    image_format=("base64" if self.image_format == "base64" else "url"),
                )
                images_bytes = self._extract_bytes_from_batch(resp_iter)

            # save images to files
            paths: List[str] = []
            for i, content in enumerate(images_bytes):
                # generate filename
                name = (
                    datetime.date.today().strftime("%Y-%m-%d") + "_" +
                    datetime.datetime.now().strftime("%H-%M-%S") + "-" +
                    self.window.core.image.make_safe_filename(self.input_prompt) + "-" +
                    str(i + 1) + ".jpg"
                )
                path = os.path.join(self.window.core.config.get_user_dir("img"), name)
                self.signals.status.emit(trans('img.status.downloading') + f" ({i + 1} / {len(images_bytes)}) -> {path}")

                if self.window.core.image.save_image(path, content):
                    paths.append(path)

            if self.inline:
                self.signals.finished_inline.emit(self.ctx, paths, self.input_prompt)
            else:
                self.signals.finished.emit(self.ctx, paths, self.input_prompt)

        except Exception as e:
            self.signals.error.emit(e)
        finally:
            self._cleanup()

    # ---------- SDK response helpers ----------

    def _extract_bytes_from_single(self, resp) -> List[bytes]:
        """
        Normalize single-image SDK response to a list of bytes.
        Accepts:
        - resp.image -> bytes or base64 str (docs say raw bytes)
        - resp.url -> download
        - dict-like/legacy: {'b64_json': ..., 'url': ...}
        """
        out: List[bytes] = []
        try:
            # preferred path: raw bytes when image_format="base64"
            img_bytes = getattr(resp, "image", None)
            if isinstance(img_bytes, (bytes, bytearray)):
                out.append(bytes(img_bytes))
                return out
            if isinstance(img_bytes, str):
                try:
                    out.append(base64.b64decode(img_bytes))
                    return out
                except Exception:
                    pass

            # url fallback
            url = getattr(resp, "url", None)
            if isinstance(url, str) and url:
                try:
                    r = requests.get(url, timeout=60)
                    if r.status_code == 200:
                        out.append(r.content)
                        return out
                except Exception:
                    pass

            # dict-like fallbacks
            if isinstance(resp, dict):
                if "b64_json" in resp and resp["b64_json"]:
                    try:
                        out.append(base64.b64decode(resp["b64_json"]))
                        return out
                    except Exception:
                        pass
                if "url" in resp and resp["url"]:
                    try:
                        r = requests.get(resp["url"], timeout=60)
                        if r.status_code == 200:
                            out.append(r.content)
                            return out
                    except Exception:
                        pass
        except Exception:
            pass
        return out

    def _extract_bytes_from_batch(self, resp_iter: Iterable) -> List[bytes]:
        """
        Normalize batch SDK response (iterable of images) to a list of bytes.
        Handles item.image (bytes/str), item.url, dict-like or bytes directly.
        """
        out: List[bytes] = []
        if resp_iter is None:
            return out
        try:
            for item in resp_iter:
                # bytes directly
                if isinstance(item, (bytes, bytearray)):
                    out.append(bytes(item))
                    continue

                # preferred: raw bytes in item.image
                img_bytes = getattr(item, "image", None)
                if isinstance(img_bytes, (bytes, bytearray)):
                    out.append(bytes(img_bytes))
                    continue
                if isinstance(img_bytes, str):
                    try:
                        out.append(base64.b64decode(img_bytes))
                        continue
                    except Exception:
                        pass

                # url fallback
                url = getattr(item, "url", None)
                if isinstance(url, str) and url:
                    try:
                        r = requests.get(url, timeout=60)
                        if r.status_code == 200:
                            out.append(r.content)
                            continue
                    except Exception:
                        pass

                # dict-like fallbacks
                if isinstance(item, dict):
                    if "b64_json" in item and item["b64_json"]:
                        try:
                            out.append(base64.b64decode(item["b64_json"]))
                            continue
                        except Exception:
                            pass
                    if "url" in item and item["url"]:
                        try:
                            r = requests.get(item["url"], timeout=60)
                            if r.status_code == 200:
                                out.append(r.content)
                                continue
                        except Exception:
                            pass
        except Exception:
            pass
        return out

    def _cleanup(self):
        """Cleanup signals to avoid multiple calls."""
        sig = self.signals
        self.signals = None
        if sig is not None:
            try:
                sig.deleteLater()
            except RuntimeError:
                pass

    # ---------- prompt utilities ----------

    @staticmethod
    def _merge_negative_prompt(prompt: str, negative: Optional[str]) -> str:
        """
        Append a negative prompt to the main text prompt for providers without a native negative_prompt field.
        """
        base = (prompt or "").strip()
        neg = (negative or "").strip()
        if not neg:
            return base
        return (base + ("\n" if base else "") + f"Negative prompt: {neg}").strip()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.07.30 00:00:00                  #
# ================================================== #

import os

from openai import AsyncOpenAI
from agents import (
    Model,
    ModelProvider,
    OpenAIChatCompletionsModel,
    set_tracing_disabled,
)

from pygpt_net.core.types import MODE_AGENT_OPENAI
from pygpt_net.item.model import ModelItem


class CustomModelProvider(ModelProvider):
    def __init__(self, client: AsyncOpenAI):
        """
        Custom Model Provider for OpenAI agents

        :param client: AsyncOpenAI client instance
        """
        super().__init__()
        self.client = client
        set_tracing_disabled(True)

    def get_model(self, model_name: str | None) -> Model:
        return OpenAIChatCompletionsModel(model=model_name, openai_client=self.client)


def get_custom_model_provider(window, model: ModelItem) -> CustomModelProvider:
    """
    Return custom OpenAI client for the specified model

    :param window: Window instance
    :param model: ModelItem instance
    :return: CustomModelProvider instance
    """
    args = window.core.models.prepare_client_args(MODE_AGENT_OPENAI, model)
    client = AsyncOpenAI(**args)
    return CustomModelProvider(client)

def get_client(window, model: ModelItem) -> AsyncOpenAI:
    """
    Return OpenAI client for the specified model

    :param window: Window instance
    :param model: ModelItem instance
    :return: AsyncOpenAI client instance
    """
    set_tracing_disabled(True)
    args = window.core.models.prepare_client_args(MODE_AGENT_OPENAI, model)
    return AsyncOpenAI(**args)

def set_openai_env(window):
    """
    Set OpenAI environment variables based on the window configuration

    :param window: Window instance
    """
    os.environ['OPENAI_API_KEY'] = str(window.core.config.get('api_key'))
    os.environ['OPENAI_API_BASE'] = str(window.core.config.get('api_endpoint'))
    os.environ['OPENAI_ORGANIZATION'] = str(window.core.config.get('organization_key'))


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.02 15:00:00                  #
# ================================================== #

from PySide6.QtGui import QPixmap, QIcon
from PySide6.QtWidgets import QVBoxLayout, QHBoxLayout, QWidget, QRadioButton, QPushButton, QComboBox, QScrollArea
from PySide6.QtCore import QSize

from pygpt_net.ui.widget.draw.painter import PainterWidget
from pygpt_net.ui.widget.element.labels import HelpLabel
from pygpt_net.ui.widget.option.combo import NoScrollCombo
from pygpt_net.utils import trans


class Painter:
    def __init__(self, window=None):
        """
        Painter UI

        :param window: Window instance
        """
        self.window = window
        self._initialized = False

    def init(self):
        """
        Initialize painter

        :return: QWidget
        """
        ui = self.window.ui
        nodes = ui.nodes
        common = self.window.controller.painter.common

        if getattr(ui, 'painter', None) is None:
            ui.painter = PainterWidget(self.window)

        key = 'painter.btn.brush'
        if nodes.get(key) is None:
            rb = QRadioButton(trans('painter.mode.paint'))
            rb.setChecked(True)
            rb.toggled.connect(self.window.controller.painter.common.set_brush_mode)
            nodes[key] = rb

        key = 'painter.btn.erase'
        if nodes.get(key) is None:
            rb = QRadioButton(trans('painter.mode.erase'))
            rb.toggled.connect(self.window.controller.painter.common.set_erase_mode)
            nodes[key] = rb

        key = 'painter.select.brush.size'
        if nodes.get(key) is None:
            sizes = common.get_sizes()
            cb = NoScrollCombo(self.window)
            cb.addItems(sizes)
            cb.currentTextChanged.connect(common.change_brush_size)
            cb.setMinimumContentsLength(10)
            cb.setSizeAdjustPolicy(QComboBox.AdjustToContents)
            nodes[key] = cb

        key = 'painter.select.canvas.size'
        if nodes.get(key) is None:
            canvas_sizes = common.get_canvas_sizes()
            cb = NoScrollCombo(self.window)
            cb.addItems(canvas_sizes)
            cb.setMinimumContentsLength(20)
            cb.setSizeAdjustPolicy(QComboBox.AdjustToContents)
            cb.currentTextChanged.connect(common.change_canvas_size)
            nodes[key] = cb

        key = 'painter.select.brush.color'
        if nodes.get(key) is None:
            cb = NoScrollCombo(self.window)
            cb.setIconSize(QSize(16, 16))
            colors = common.get_colors()
            for color_name, color_value in colors.items():
                pixmap = QPixmap(16, 16)
                pixmap.fill(color_value)
                icon = QIcon(pixmap)
                cb.addItem(icon, color_name, color_value)
            cb.currentTextChanged.connect(common.change_brush_color)
            cb.setMinimumContentsLength(10)
            cb.setMinimumWidth(205)
            cb.setSizeAdjustPolicy(QComboBox.AdjustToContents)
            nodes[key] = cb

            # Zoom combo (view-only scale) placed to the right of canvas size
            key = 'painter.select.zoom'
            if nodes.get(key) is None:
                cb = NoScrollCombo(self.window)
                cb.setMinimumContentsLength(8)
                cb.setSizeAdjustPolicy(QComboBox.AdjustToContents)

                # Preferred preset steps from widget; fallback to defaults
                steps = []
                if hasattr(ui.painter, 'get_zoom_steps_percent'):
                    try:
                        steps = ui.painter.get_zoom_steps_percent()
                    except Exception:
                        steps = []
                if not steps:
                    steps = [10, 25, 50, 75, 100, 150, 200, 500, 1000]

                cb.addItems([f"{p}%" for p in steps])

                # User -> widget
                cb.currentTextChanged.connect(ui.painter.on_zoom_combo_changed)

                # Widget -> combo (also covers CTRL+wheel and programmatic changes)
                def _sync_zoom_combo_from_widget(z):
                    """Keep zoom combobox in sync with the widget's zoom."""
                    percent = int(round(float(z) * 100))
                    label = f"{percent}%"
                    cb.blockSignals(True)
                    idx = cb.findText(label)
                    if idx >= 0:
                        cb.setCurrentIndex(idx)
                    else:
                        # Insert missing value keeping ascending order
                        items = [cb.itemText(i) for i in range(cb.count())]
                        if label not in items:
                            items.append(label)
                            try:
                                items_sorted = sorted(
                                    set(items),
                                    key=lambda s: float(s.replace('%', '').strip())
                                )
                            except Exception:
                                items_sorted = items
                            cb.clear()
                            cb.addItems(items_sorted)
                        cb.setCurrentText(label)
                    cb.blockSignals(False)

                # Keep reference to prevent GC of the inner function
                cb._sync_zoom_combo_from_widget = _sync_zoom_combo_from_widget
                if hasattr(ui.painter, 'zoomChanged'):
                    ui.painter.zoomChanged.connect(cb._sync_zoom_combo_from_widget)

                # Initial label; actual value will be set by load_zoom below
                cb.setCurrentText("100%")
                nodes[key] = cb

        self._initialized = True

    def setup(self) -> QWidget:
        """
        Setup painter

        :return: QWidget
        """
        self.init()
        body = self.window.core.tabs.from_layout(self.setup_painter())
        body.append(self.window.ui.painter)
        return body

    def setup_painter(self) -> QVBoxLayout:
        """
        Setup painter

        :return: QVBoxLayout
        """
        ui = self.window.ui
        nodes = ui.nodes

        top = QHBoxLayout()
        top.addWidget(nodes['painter.btn.brush'])
        top.addWidget(nodes['painter.btn.erase'])
        top.addWidget(nodes['painter.select.brush.size'])
        top.addWidget(nodes['painter.select.brush.color'])
        top.addWidget(nodes['painter.select.canvas.size'])
        # Zoom combo placed right after canvas size
        top.addWidget(nodes['painter.select.zoom'])
        top.addStretch(1)

        if nodes.get('painter.btn.capture') is None:
            btn = QPushButton(trans('painter.btn.capture'))
            btn.clicked.connect(self.window.controller.painter.capture.use)
            nodes['painter.btn.capture'] = btn
        top.addWidget(nodes['painter.btn.capture'])

        if nodes.get('painter.btn.camera.capture') is None:
            btn = QPushButton(trans('painter.btn.camera.capture'))
            btn.clicked.connect(self.window.controller.painter.capture.camera)
            nodes['painter.btn.camera.capture'] = btn
        top.addWidget(nodes['painter.btn.camera.capture'])

        if nodes.get('painter.btn.clear') is None:
            btn = QPushButton(trans('painter.btn.clear'))
            btn.clicked.connect(ui.painter.clear_image)
            nodes['painter.btn.clear'] = btn
        top.addWidget(nodes['painter.btn.clear'])

        if getattr(ui, 'painter_scroll', None) is None:
            ui.painter_scroll = QScrollArea()
            ui.painter_scroll.setWidget(ui.painter)
            # Must be False to allow content widget to grow/shrink with zoom and show scrollbars
            ui.painter_scroll.setWidgetResizable(False)
        else:
            if ui.painter_scroll.widget() is not ui.painter:
                ui.painter_scroll.setWidget(ui.painter)
            ui.painter_scroll.setWidgetResizable(False)

        if nodes.get('tip.output.tab.draw') is None:
            nodes['tip.output.tab.draw'] = HelpLabel(trans('tip.output.tab.draw'), self.window)

        layout = QVBoxLayout()
        layout.addLayout(top)
        layout.addWidget(ui.painter_scroll)
        layout.addWidget(nodes['tip.output.tab.draw'])
        layout.setContentsMargins(0, 0, 0, 0)

        return layout


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.08.22 10:00:00                  #
# ================================================== #

from __future__ import annotations

import os
import shlex
import subprocess
import ftplib
import smtplib
import time
import stat as pystat

from email.message import EmailMessage
from PySide6.QtCore import Slot

from pygpt_net.plugin.base.worker import BaseWorker, BaseSignals


class WorkerSignals(BaseSignals):
    pass


class Worker(BaseWorker):
    """
    Server plugin worker: SSH, SFTP, FTP, Telnet, SMTP.
    Credentials are fetched via self.get_server_config(server, port).
    When prefer_system_ssh is enabled, uses system ssh/scp/sftp and system keys.
    """

    # Global map SERVICE -> PORT (defaults)
    SERVICE_PORTS = {
        "ssh": 22,
        "sftp": 22,
        "ftp": 21,
        "telnet": 23,
        "smtp": 25,
        "smtps": 465,
        "submission": 587,
    }

    def __init__(self, *args, **kwargs):
        super(Worker, self).__init__()
        self.signals = WorkerSignals()
        self.args = args
        self.kwargs = kwargs
        self.plugin = None
        self.cmds = None
        self.ctx = None
        self.msg = None

    # ---------------------- Runner ----------------------

    @Slot()
    def run(self):
        try:
            responses = []
            for item in self.cmds:
                if self.is_stopped():
                    break
                try:
                    response = None
                    if item["cmd"] in self.plugin.allowed_cmds and self.plugin.has_cmd(item["cmd"]):

                        # -------- Core FS / Exec --------
                        if item["cmd"] == "srv_exec":
                            response = self.cmd_srv_exec(item)
                        elif item["cmd"] == "srv_ls":
                            response = self.cmd_srv_ls(item)
                        elif item["cmd"] == "srv_get":
                            response = self.cmd_srv_get(item)
                        elif item["cmd"] == "srv_put":
                            response = self.cmd_srv_put(item)
                        elif item["cmd"] == "srv_rm":
                            response = self.cmd_srv_rm(item)
                        elif item["cmd"] == "srv_mkdir":
                            response = self.cmd_srv_mkdir(item)
                        elif item["cmd"] == "srv_stat":
                            response = self.cmd_srv_stat(item)

                        # -------- SMTP --------
                        elif item["cmd"] == "smtp_send":
                            response = self.cmd_smtp_send(item)

                        if response:
                            responses.append(response)

                except Exception as e:
                    responses.append(self.make_response(item, self.throw_error(e)))

            if responses:
                self.reply_more(responses)
            if self.msg is not None:
                self.status(self.msg)
        except Exception as e:
            self.error(e)
        finally:
            self.cleanup()

    # ---------------------- Common helpers ----------------------

    def _timeout(self) -> int:
        try:
            return int(self.plugin.get_option_value("net_timeout") or 30)
        except Exception:
            return 30

    def prepare_path(self, path: str) -> str:
        if path in [".", "./"]:
            return self.plugin.window.core.config.get_user_dir("data")
        if self.is_absolute_path(path):
            return path
        return os.path.join(self.plugin.window.core.config.get_user_dir("data"), path)

    def is_absolute_path(self, path: str) -> bool:
        return os.path.isabs(path)

    def _prefer_system_ssh(self) -> bool:
        return bool(self.plugin.get_option_value("prefer_system_ssh") or False)

    def _ssh_bins(self) -> dict:
        return {
            "ssh": (self.plugin.get_option_value("ssh_binary") or "ssh").strip(),
            "scp": (self.plugin.get_option_value("scp_binary") or "scp").strip(),
            "sftp": (self.plugin.get_option_value("sftp_binary") or "sftp").strip(),
        }

    def _ssh_options(self) -> list[str]:
        # Always enforce BatchMode=yes to avoid hanging on password prompt
        extra = (self.plugin.get_option_value("ssh_options") or "").strip()
        opts = ["-o", "BatchMode=yes"]
        if extra:
            opts += shlex.split(extra)
        return opts

    def _ssh_auto_add_hostkey(self) -> bool:
        return bool(self.plugin.get_option_value("ssh_auto_add_hostkey") or True)

    def _ftp_use_tls_default(self) -> bool:
        return bool(self.plugin.get_option_value("ftp_use_tls_default") or False)

    def _ftp_passive_default(self) -> bool:
        return bool(self.plugin.get_option_value("ftp_passive_default") or True)

    def _smtp_defaults(self) -> dict:
        return {
            "use_tls": bool(self.plugin.get_option_value("smtp_use_tls_default") or True),
            "use_ssl": bool(self.plugin.get_option_value("smtp_use_ssl_default") or False),
            "from_addr": (self.plugin.get_option_value("smtp_from_default") or "").strip(),
        }

    def _ensure_paramiko(self):
        try:
            import paramiko  # noqa
        except Exception:
            raise RuntimeError("Paramiko not installed. Install: pip install paramiko")

    def _service_from_port(self, port: int) -> str:
        for k, v in self.SERVICE_PORTS.items():
            if int(v) == int(port):
                return k
        return "unknown"

    def _server_config(self, server: str, port: int) -> dict:
        cfg = self.get_server_config(server, int(port))  # raises RuntimeError with available server names if not found
        if not cfg.get("server") or not cfg.get("login"):
            raise RuntimeError("Server config incomplete (server/login).")
        # Never expose credentials
        return cfg

    def _shell_quote(self, s: str) -> str:
        return shlex.quote(str(s))

    def _build_remote_cmd(self, command: str, cwd: str | None = None, env: dict | None = None) -> str:
        parts = []
        if env:
            exports = " ".join([f'{k}={self._shell_quote(v)}' for k, v in env.items()])
            parts.append(f"export {exports}")
        if cwd:
            parts.append(f"cd {self._shell_quote(cwd)}")
        parts.append(command)
        joined = " && ".join(parts)
        # Use bash -lc to ensure login-like shell behaviors (PATH, etc.)
        return f"bash -lc {self._shell_quote(joined)}"

    def get_server_config(self, server: str, port: int) -> dict:
        """
        Get server configuration for given server and port.

        :param server: server name or host
        :param port: server port
        :return: dict with server configuration (server, login, password, port)
        """
        servers = self.plugin.get_option_value("servers")
        available_ids = []
        for srv in servers:
            if not srv.get("enabled"):
                continue
            available_ids.append(srv.get("name"))
        config = {}
        for srv in servers:
            if not srv.get("enabled"):
                continue
            if ((srv.get("name").lower() == server.lower()
                 or srv.get("host").lower() == server.lower())
                    and int(srv.get("port", 0)) == port):
                config = {
                    "server": srv.get("host"),
                    "login": srv.get("login"),
                    "password": srv.get("password"),
                    "port": int(srv.get("port", 0)),
                }
                break
        if not config:
            raise RuntimeError(
                f"Server configuration not found for {server}:{port}. Available servers: {', '.join(available_ids)}")
        return config

    # ---------------------- SSH (system) ----------------------

    def _ssh_exec_system(self, host: str, port: int, user: str, command: str, cwd: str | None,
                         env: dict | None) -> dict:
        bins = self._ssh_bins()
        base = [bins["ssh"]] + self._ssh_options() + ["-p", str(port)]
        remote = f"{user}@{host}"
        rcmd = self._build_remote_cmd(command, cwd=cwd, env=env)
        try:
            r = subprocess.run(base + [remote, rcmd], capture_output=True, text=True, timeout=self._timeout())
            return {"rc": r.returncode, "stdout": r.stdout, "stderr": r.stderr}
        except subprocess.TimeoutExpired as e:
            raise RuntimeError(f"SSH timeout after {self._timeout()}s: {e}")
        except FileNotFoundError:
            raise RuntimeError("ssh binary not found.")

    def _scp_get_system(self, host: str, port: int, user: str, remote_path: str, local_path: str) -> dict:
        bins = self._ssh_bins()
        base = [bins["scp"]] + self._ssh_options() + ["-P", str(port)]
        remote = f"{user}@{host}:{remote_path}"
        os.makedirs(os.path.dirname(local_path) or ".", exist_ok=True)
        try:
            r = subprocess.run(base + [remote, local_path], capture_output=True, text=True, timeout=self._timeout())
            if r.returncode != 0:
                raise RuntimeError(f"SCP get failed: {r.stderr.strip()}")
            size = os.path.getsize(local_path) if os.path.exists(local_path) else None
            return {"saved_path": local_path, "size": size}
        except subprocess.TimeoutExpired as e:
            raise RuntimeError(f"SCP timeout after {self._timeout()}s: {e}")
        except FileNotFoundError:
            raise RuntimeError("scp binary not found.")

    def _scp_put_system(self, host: str, port: int, user: str, local_path: str, remote_path: str) -> dict:
        bins = self._ssh_bins()
        base = [bins["scp"]] + self._ssh_options() + ["-P", str(port)]
        if not os.path.exists(local_path):
            raise RuntimeError(f"Local path not found: {local_path}")
        remote = f"{user}@{host}:{remote_path}"
        try:
            r = subprocess.run(base + [local_path, remote], capture_output=True, text=True, timeout=self._timeout())
            if r.returncode != 0:
                raise RuntimeError(f"SCP put failed: {r.stderr.strip()}")
            return {"uploaded": True, "source": local_path, "dest": remote_path}
        except subprocess.TimeoutExpired as e:
            raise RuntimeError(f"SCP timeout after {self._timeout()}s: {e}")
        except FileNotFoundError:
            raise RuntimeError("scp binary not found.")

    def _ssh_ls_system(self, host: str, port: int, user: str, path: str | None) -> dict:
        # Best-effort parse of `ls -la`. Not perfect on all systems.
        p = path or "."
        cmd = f'ls -la {self._shell_quote(p)}'
        res = self._ssh_exec_system(host, port, user, cmd, cwd=None, env=None)
        if res["rc"] != 0:
            raise RuntimeError(f"ls failed: {res['stderr'].strip()}")
        entries = []
        for line in res["stdout"].splitlines():
            if not line or line.startswith("total "):
                continue
            parts = line.split(maxsplit=8)
            if len(parts) < 9:
                continue
            perms, _, owner, group, size, month, day, time_or_year, name = parts
            if name in [".", ".."]:
                continue
            ftype = "file"
            if perms.startswith("d"):
                ftype = "dir"
            elif perms.startswith("l"):
                ftype = "link"
            # size may fail to int if localized
            try:
                size_val = int(size)
            except Exception:
                size_val = None
            entries.append({"name": name, "type": ftype, "size": size_val, "raw": line})
        return {"entries": entries, "raw": res["stdout"]}

    # ---------------------- SSH / SFTP (paramiko) ----------------------

    def _ssh_exec_paramiko(self, host: str, port: int, user: str, password: str | None, command: str, cwd: str | None,
                           env: dict | None) -> dict:
        self._ensure_paramiko()
        import paramiko
        client = paramiko.SSHClient()
        if self._ssh_auto_add_hostkey():
            client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        try:
            client.connect(
                host,
                port=port,
                username=user,
                password=password or None,
                timeout=self._timeout(),
                allow_agent=True,
                look_for_keys=True,
            )
            rcmd = self._build_remote_cmd(command, cwd=cwd, env=env)
            stdin, stdout, stderr = client.exec_command(rcmd, timeout=self._timeout())
            out = stdout.read().decode("utf-8", errors="replace")
            err = stderr.read().decode("utf-8", errors="replace")
            rc = stdout.channel.recv_exit_status()
            return {"rc": rc, "stdout": out, "stderr": err}
        finally:
            try:
                client.close()
            except Exception:
                pass

    def _sftp_opener(self, host: str, port: int, user: str, password: str | None):
        self._ensure_paramiko()
        import paramiko
        client = paramiko.SSHClient()
        if self._ssh_auto_add_hostkey():
            client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        client.connect(
            host,
            port=port,
            username=user,
            password=password or None,
            timeout=self._timeout(),
            allow_agent=True,
            look_for_keys=True,
        )
        sftp = client.open_sftp()
        return client, sftp

    def _sftp_ls(self, host: str, port: int, user: str, password: str | None, path: str | None) -> dict:
        client, sftp = self._sftp_opener(host, port, user, password)
        try:
            p = path or "."
            attrs = sftp.listdir_attr(p)
            out = []
            for a in attrs:
                name = getattr(a, "filename", None)
                mode = a.st_mode if hasattr(a, "st_mode") else 0
                ftype = "file"
                if pystat.S_ISDIR(mode):
                    ftype = "dir"
                elif pystat.S_ISLNK(mode):
                    ftype = "link"
                out.append({
                    "name": name,
                    "type": ftype,
                    "size": getattr(a, "st_size", None),
                    "mtime": getattr(a, "st_mtime", None),
                    "mode": mode,
                })
            return {"entries": out}
        finally:
            try:
                sftp.close()
            except Exception:
                pass
            try:
                client.close()
            except Exception:
                pass

    def _sftp_get(self, host: str, port: int, user: str, password: str | None, remote_path: str,
                  local_path: str) -> dict:
        client, sftp = self._sftp_opener(host, port, user, password)
        try:
            os.makedirs(os.path.dirname(local_path) or ".", exist_ok=True)
            sftp.get(remote_path, local_path)
            size = os.path.getsize(local_path)
            return {"saved_path": local_path, "size": size}
        finally:
            try:
                sftp.close()
            except Exception:
                pass
            try:
                client.close()
            except Exception:
                pass

    def _sftp_put(self, host: str, port: int, user: str, password: str | None, local_path: str, remote_path: str,
                  make_dirs: bool) -> dict:
        if not os.path.exists(local_path):
            raise RuntimeError(f"Local path not found: {local_path}")
        client, sftp = self._sftp_opener(host, port, user, password)
        try:
            if make_dirs:
                # naive mkdir -p
                dir_part = os.path.dirname(remote_path)
                if dir_part:
                    self._sftp_mkdirs(sftp, dir_part)
            sftp.put(local_path, remote_path)
            return {"uploaded": True, "source": local_path, "dest": remote_path}
        finally:
            try:
                sftp.close()
            except Exception:
                pass
            try:
                client.close()
            except Exception:
                pass

    def _sftp_mkdirs(self, sftp, remote_dir: str):
        parts = []
        while True:
            head, tail = os.path.split(remote_dir)
            if tail:
                parts.append(tail)
                remote_dir = head
            else:
                if head:
                    parts.append(head)
                break
        parts = list(reversed([p for p in parts if p not in ["", "/"]]))
        cur = ""
        for p in parts:
            if cur == "":
                if remote_dir.startswith("/"):
                    cur = f"/{p}"
                else:
                    cur = p
            else:
                cur = f"{cur}/{p}"
            try:
                sftp.stat(cur)
            except Exception:
                try:
                    sftp.mkdir(cur)
                except Exception:
                    pass

    def _sftp_rm(self, host: str, port: int, user: str, password: str | None, remote_path: str) -> dict:
        client, sftp = self._sftp_opener(host, port, user, password)
        try:
            st = sftp.stat(remote_path)
            if pystat.S_ISDIR(st.st_mode):
                # Non-recursive rmdir
                sftp.rmdir(remote_path)
                return {"removed": True, "path": remote_path, "type": "dir"}
            else:
                sftp.remove(remote_path)
                return {"removed": True, "path": remote_path, "type": "file"}
        finally:
            try:
                sftp.close()
            except Exception:
                pass
            try:
                client.close()
            except Exception:
                pass

    def _sftp_mkdir(self, host: str, port: int, user: str, password: str | None, remote_path: str,
                    exist_ok: bool) -> dict:
        client, sftp = self._sftp_opener(host, port, user, password)
        try:
            try:
                sftp.mkdir(remote_path)
            except IOError:
                if not exist_ok:
                    raise
            return {"mkdir": True, "path": remote_path}
        finally:
            try:
                sftp.close()
            except Exception:
                pass
            try:
                client.close()
            except Exception:
                pass

    def _sftp_stat(self, host: str, port: int, user: str, password: str | None, remote_path: str) -> dict:
        client, sftp = self._sftp_opener(host, port, user, password)
        try:
            st = sftp.stat(remote_path)
            mode = st.st_mode
            ftype = "file"
            if pystat.S_ISDIR(mode):
                ftype = "dir"
            elif pystat.S_ISLNK(mode):
                ftype = "link"
            return {
                "path": remote_path,
                "type": ftype,
                "size": getattr(st, "st_size", None),
                "mtime": getattr(st, "st_mtime", None),
                "mode": mode,
            }
        finally:
            try:
                sftp.close()
            except Exception:
                pass
            try:
                client.close()
            except Exception:
                pass

    # ---------------------- FTP / FTPS (stdlib) ----------------------

    def _ftp_connect(self, host: str, port: int, user: str, password: str | None, use_tls: bool, passive: bool):
        if use_tls:
            ftp = ftplib.FTP_TLS()
        else:
            ftp = ftplib.FTP()
        ftp.connect(host, port, timeout=self._timeout())
        ftp.login(user=user, passwd=password or "")
        if use_tls:
            try:
                ftp.prot_p()  # secure data channel
            except Exception:
                pass
        ftp.set_pasv(passive)
        return ftp

    def _ftp_ls(self, host: str, port: int, user: str, password: str | None, path: str | None) -> dict:
        use_tls = self._ftp_use_tls_default() or (int(port) in [990])
        passive = self._ftp_passive_default()
        ftp = self._ftp_connect(host, port, user, password, use_tls, passive)
        try:
            p = path or "."
            entries = []
            # Prefer MLSD if supported
            try:
                for name, facts in ftp.mlsd(p):
                    ftype = facts.get("type", "file")
                    if ftype == "dir":
                        t = "dir"
                    elif ftype == "file":
                        t = "file"
                    else:
                        t = ftype
                    size = None
                    try:
                        size = int(facts.get("size")) if "size" in facts else None
                    except Exception:
                        pass
                    mtime = None
                    try:
                        # facts['modify'] like YYYYMMDDHHMMSS
                        mod = facts.get("modify")
                        if mod and len(mod) >= 14:
                            # convert to epoch best-effort
                            mtime = int(time.mktime(time.strptime(mod, "%Y%m%d%H%M%S")))
                    except Exception:
                        pass
                    entries.append({"name": name, "type": t, "size": size, "mtime": mtime})
            except Exception:
                # Fallback to NLST (names only)
                names = ftp.nlst(p)
                for name in names:
                    base = os.path.basename(name)
                    entries.append({"name": base, "type": "unknown"})
            return {"entries": entries}
        finally:
            try:
                ftp.quit()
            except Exception:
                pass

    def _ftp_get(self, host: str, port: int, user: str, password: str | None, remote_path: str,
                 local_path: str) -> dict:
        use_tls = self._ftp_use_tls_default() or (int(port) in [990])
        passive = self._ftp_passive_default()
        ftp = self._ftp_connect(host, port, user, password, use_tls, passive)
        try:
            os.makedirs(os.path.dirname(local_path) or ".", exist_ok=True)
            with open(local_path, "wb") as fh:
                ftp.retrbinary(f"RETR {remote_path}", fh.write)
            size = os.path.getsize(local_path)
            return {"saved_path": local_path, "size": size}
        finally:
            try:
                ftp.quit()
            except Exception:
                pass

    def _ftp_put(self, host: str, port: int, user: str, password: str | None, local_path: str,
                 remote_path: str) -> dict:
        if not os.path.exists(local_path):
            raise RuntimeError(f"Local path not found: {local_path}")
        use_tls = self._ftp_use_tls_default() or (int(port) in [990])
        passive = self._ftp_passive_default()
        ftp = self._ftp_connect(host, port, user, password, use_tls, passive)
        try:
            with open(local_path, "rb") as fh:
                ftp.storbinary(f"STOR {remote_path}", fh)
            return {"uploaded": True, "source": local_path, "dest": remote_path}
        finally:
            try:
                ftp.quit()
            except Exception:
                pass

    def _ftp_rm(self, host: str, port: int, user: str, password: str | None, remote_path: str) -> dict:
        use_tls = self._ftp_use_tls_default() or (int(port) in [990])
        passive = self._ftp_passive_default()
        ftp = self._ftp_connect(host, port, user, password, use_tls, passive)
        try:
            try:
                ftp.delete(remote_path)
                return {"removed": True, "path": remote_path, "type": "file"}
            except ftplib.error_perm:
                # maybe it's a dir (non-recursive)
                ftp.rmd(remote_path)
                return {"removed": True, "path": remote_path, "type": "dir"}
        finally:
            try:
                ftp.quit()
            except Exception:
                pass

    def _ftp_mkdir(self, host: str, port: int, user: str, password: str | None, remote_path: str,
                   exist_ok: bool) -> dict:
        use_tls = self._ftp_use_tls_default() or (int(port) in [990])
        passive = self._ftp_passive_default()
        ftp = self._ftp_connect(host, port, user, password, use_tls, passive)
        try:
            try:
                ftp.mkd(remote_path)
            except ftplib.error_perm:
                if not exist_ok:
                    raise
            return {"mkdir": True, "path": remote_path}
        finally:
            try:
                ftp.quit()
            except Exception:
                pass

    def _ftp_stat(self, host: str, port: int, user: str, password: str | None, remote_path: str) -> dict:
        use_tls = self._ftp_use_tls_default() or (int(port) in [990])
        passive = self._ftp_passive_default()
        ftp = self._ftp_connect(host, port, user, password, use_tls, passive)
        try:
            # Try MLST single path
            try:
                resp = []
                ftp.sendcmd("TYPE I")  # binary
                ftp.sendcmd(f"MLST {remote_path}")
                # ftplib doesn't parse MLST easily; use voidcmd and handle later is messy.
            except Exception:
                pass
            # Fallback: list parent and find entry
            parent = os.path.dirname(remote_path) or "."
            name = os.path.basename(remote_path)
            info = None
            try:
                for n, facts in ftp.mlsd(parent):
                    if n == name:
                        info = facts
                        break
            except Exception:
                pass
            if info:
                t = info.get("type", "file")
                ftype = "dir" if t == "dir" else ("file" if t == "file" else t)
                size = None
                try:
                    size = int(info.get("size")) if "size" in info else None
                except Exception:
                    pass
                mtime = None
                try:
                    mod = info.get("modify")
                    if mod and len(mod) >= 14:
                        mtime = int(time.mktime(time.strptime(mod, "%Y%m%d%H%M%S")))
                except Exception:
                    pass
                return {"path": remote_path, "type": ftype, "size": size, "mtime": mtime}
            # As last resort, try size cmd (files only)
            try:
                ftp.sendcmd("TYPE I")
                size = ftp.size(remote_path)
                return {"path": remote_path, "type": "file", "size": size}
            except Exception:
                pass
            # Unknown
            return {"path": remote_path, "type": "unknown"}
        finally:
            try:
                ftp.quit()
            except Exception:
                pass

    # ---------------------- Telnet (stdlib, best-effort) ----------------------

    def _telnet_exec(self, host: str, port: int, user: str, password: str | None, command: str, login_prompt: str,
                     password_prompt: str, prompt: str) -> dict:
        try:
            import telnetlib  # deprecated in newer Python, but still available in many
        except Exception:
            raise RuntimeError("telnetlib not available. Use SSH instead.")
        timeout = self._timeout()
        tn = telnetlib.Telnet(host, port, timeout=timeout)
        try:
            if login_prompt:
                tn.read_until(login_prompt.encode("utf-8"), timeout=timeout)
                tn.write((user + "\n").encode("utf-8"))
            if password_prompt is not None:
                tn.read_until(password_prompt.encode("utf-8"), timeout=timeout)
                tn.write(((password or "") + "\n").encode("utf-8"))
            if prompt:
                tn.read_until(prompt.encode("utf-8"), timeout=timeout)
            tn.write((command + "\n").encode("utf-8"))
            out = tn.read_until(prompt.encode("utf-8"), timeout=timeout)
            # crude strip
            text = out.decode("utf-8", errors="replace")
            return {"rc": 0, "stdout": text, "stderr": ""}
        finally:
            try:
                tn.close()
            except Exception:
                pass

    # ---------------------- SMTP (stdlib) ----------------------

    def _smtp_send_mail(self, host: str, port: int, user: str, password: str | None, mail: dict) -> dict:
        defaults = self._smtp_defaults()
        use_tls = bool(mail.get("use_tls", defaults["use_tls"]))
        use_ssl = bool(mail.get("use_ssl", defaults["use_ssl"]))
        from_addr = (mail.get("from_addr") or defaults["from_addr"] or user).strip()
        to = mail.get("to")
        if isinstance(to, str):
            to = [to]
        cc = mail.get("cc") or []
        bcc = mail.get("bcc") or []
        subject = mail.get("subject") or ""
        body = mail.get("body") or ""
        html = bool(mail.get("html") or False)

        if not to:
            raise RuntimeError("Param 'to' required for smtp_send.")

        msg = EmailMessage()
        msg["From"] = from_addr
        msg["To"] = ", ".join(to)
        if cc:
            msg["Cc"] = ", ".join(cc)
        msg["Subject"] = subject
        if html:
            msg.add_alternative(body, subtype="html")
        else:
            msg.set_content(body)

        # Attachments (local files)
        for apath in (mail.get("attachments") or []):
            lp = self.prepare_path(apath)
            if not os.path.exists(lp):
                raise RuntimeError(f"Attachment not found: {lp}")
            with open(lp, "rb") as f:
                data = f.read()
            maintype, subtype = "application", "octet-stream"
            msg.add_attachment(data, maintype=maintype, subtype=subtype, filename=os.path.basename(lp))

        if use_ssl or int(port) in [465]:
            smtp = smtplib.SMTP_SSL(host, int(port), timeout=self._timeout())
        else:
            smtp = smtplib.SMTP(host, int(port), timeout=self._timeout())
        try:
            smtp.ehlo()
            if use_tls and not use_ssl:
                smtp.starttls()
                smtp.ehlo()
            if user:
                smtp.login(user, password or "")
            rcpt = to + (cc if isinstance(cc, list) else []) + (bcc if isinstance(bcc, list) else [])
            smtp.send_message(msg, from_addr=from_addr, to_addrs=rcpt)
            return {"sent": True, "to": rcpt}
        finally:
            try:
                smtp.quit()
            except Exception:
                pass

    # ---------------------- Commands ----------------------

    def cmd_srv_exec(self, item: dict) -> dict:
        p = item.get("params", {}) or {}
        server = p.get("server")
        port = int(p.get("port") or 0)
        command = p.get("command")
        cwd = p.get("cwd")
        env = p.get("env") or {}
        if not (server and port and command):
            return self.make_response(item, "Params 'server', 'port' and 'command' required.")

        cfg = self._server_config(server, port)
        host = cfg.get("server")
        user = cfg.get("login")
        password = cfg.get("password")

        svc = self._service_from_port(port)
        if svc in ["ssh", "sftp", "unknown"] or int(port) == 22:
            if self._prefer_system_ssh():
                res = self._ssh_exec_system(host, port, user, command, cwd=cwd, env=env)
                return self.make_response(item, res)
            else:
                res = self._ssh_exec_paramiko(host, port, user, password, command, cwd=cwd, env=env)
                return self.make_response(item, res)
        elif svc == "telnet" or int(port) == 23:
            login_prompt = (self.plugin.get_option_value("telnet_login_prompt") or "login:").strip()
            password_prompt = (self.plugin.get_option_value("telnet_password_prompt") or "Password:").strip()
            prompt = (self.plugin.get_option_value("telnet_prompt") or "$ ").strip()
            res = self._telnet_exec(host, port, user, password, command, login_prompt, password_prompt, prompt)
            return self.make_response(item, res)
        else:
            return self.make_response(item, f"srv_exec not supported for port {port} ({svc}).")

    def cmd_srv_ls(self, item: dict) -> dict:
        p = item.get("params", {}) or {}
        server = p.get("server")
        port = int(p.get("port") or 0)
        path = p.get("path") or "."
        if not (server and port):
            return self.make_response(item, "Params 'server' and 'port' required.")
        cfg = self._server_config(server, port)
        host = cfg.get("server")
        user = cfg.get("login")
        password = cfg.get("password")
        svc = self._service_from_port(port)
        if svc in ["ssh", "sftp", "unknown"] or int(port) == 22:
            if self._prefer_system_ssh():
                res = self._ssh_ls_system(host, port, user, path)
                return self.make_response(item, res)
            else:
                res = self._sftp_ls(host, port, user, password, path)
                return self.make_response(item, res)
        elif svc == "ftp" or int(port) in [21, 990]:
            res = self._ftp_ls(host, port, user, password, path)
            return self.make_response(item, res)
        else:
            return self.make_response(item, f"srv_ls not supported for port {port} ({svc}).")

    def cmd_srv_get(self, item: dict) -> dict:
        p = item.get("params", {}) or {}
        server = p.get("server")
        port = int(p.get("port") or 0)
        remote_path = p.get("remote_path")
        local_path = p.get("local_path")
        overwrite = bool(p.get("overwrite", True))
        if not (server and port and remote_path):
            return self.make_response(item, "Params 'server', 'port' and 'remote_path' required.")
        if not local_path:
            local_path = self.prepare_path(os.path.basename(remote_path))
        else:
            local_path = self.prepare_path(local_path)
        if os.path.exists(local_path) and not overwrite:
            return self.make_response(item, f"Local path exists and overwrite=False: {local_path}")

        cfg = self._server_config(server, port)
        host = cfg.get("server")
        user = cfg.get("login")
        password = cfg.get("password")
        svc = self._service_from_port(port)
        if svc in ["ssh", "sftp", "unknown"] or int(port) == 22:
            if self._prefer_system_ssh():
                res = self._scp_get_system(host, port, user, remote_path, local_path)
                return self.make_response(item, res)
            else:
                res = self._sftp_get(host, port, user, password, remote_path, local_path)
                return self.make_response(item, res)
        elif svc == "ftp" or int(port) in [21, 990]:
            res = self._ftp_get(host, port, user, password, remote_path, local_path)
            return self.make_response(item, res)
        else:
            return self.make_response(item, f"srv_get not supported for port {port} ({svc}).")

    def cmd_srv_put(self, item: dict) -> dict:
        p = item.get("params", {}) or {}
        server = p.get("server")
        port = int(p.get("port") or 0)
        local_path = p.get("local_path")
        remote_path = p.get("remote_path")
        make_dirs = bool(p.get("make_dirs", True))
        if not (server and port and local_path):
            return self.make_response(item, "Params 'server', 'port' and 'local_path' required.")
        lp = self.prepare_path(local_path)
        if not os.path.exists(lp):
            return self.make_response(item, f"Local path not found: {lp}")
        if not remote_path:
            remote_path = os.path.basename(lp)

        cfg = self._server_config(server, port)
        host = cfg.get("server")
        user = cfg.get("login")
        password = cfg.get("password")
        svc = self._service_from_port(port)
        if svc in ["ssh", "sftp", "unknown"] or int(port) == 22:
            if self._prefer_system_ssh():
                # System scp cannot create remote dirs automatically; rely on user setting up path
                res = self._scp_put_system(host, port, user, lp, remote_path)
                return self.make_response(item, res)
            else:
                res = self._sftp_put(host, port, user, password, lp, remote_path, make_dirs=make_dirs)
                return self.make_response(item, res)
        elif svc == "ftp" or int(port) in [21, 990]:
            res = self._ftp_put(host, port, user, password, lp, remote_path)
            return self.make_response(item, res)
        else:
            return self.make_response(item, f"srv_put not supported for port {port} ({svc}).")

    def cmd_srv_rm(self, item: dict) -> dict:
        p = item.get("params", {}) or {}
        server = p.get("server")
        port = int(p.get("port") or 0)
        remote_path = p.get("remote_path")
        if not (server and port and remote_path):
            return self.make_response(item, "Params 'server', 'port' and 'remote_path' required.")
        cfg = self._server_config(server, port)
        host = cfg.get("server")
        user = cfg.get("login")
        password = cfg.get("password")
        svc = self._service_from_port(port)
        if svc in ["ssh", "sftp", "unknown"] or int(port) == 22:
            if self._prefer_system_ssh():
                # Use rm/rmdir via SSH (non-recursive)
                cmd = f"if [ -d {self._shell_quote(remote_path)} ]; then rmdir {self._shell_quote(remote_path)}; else rm -f {self._shell_quote(remote_path)}; fi"
                res = self._ssh_exec_system(host, port, user, cmd, cwd=None, env=None)
                if res["rc"] != 0:
                    raise RuntimeError(f"remove failed: {res['stderr'].strip()}")
                return self.make_response(item, {"removed": True, "path": remote_path})
            else:
                res = self._sftp_rm(host, port, user, password, remote_path)
                return self.make_response(item, res)
        elif svc == "ftp" or int(port) in [21, 990]:
            res = self._ftp_rm(host, port, user, password, remote_path)
            return self.make_response(item, res)
        else:
            return self.make_response(item, f"srv_rm not supported for port {port} ({svc}).")

    def cmd_srv_mkdir(self, item: dict) -> dict:
        p = item.get("params", {}) or {}
        server = p.get("server")
        port = int(p.get("port") or 0)
        remote_path = p.get("remote_path")
        exist_ok = bool(p.get("exist_ok", True))
        if not (server and port and remote_path):
            return self.make_response(item, "Params 'server', 'port' and 'remote_path' required.")
        cfg = self._server_config(server, port)
        host = cfg.get("server")
        user = cfg.get("login")
        password = cfg.get("password")
        svc = self._service_from_port(port)
        if svc in ["ssh", "sftp", "unknown"] or int(port) == 22:
            if self._prefer_system_ssh():
                cmd = f"mkdir {'-p ' if exist_ok else ''}{self._shell_quote(remote_path)}"
                res = self._ssh_exec_system(host, port, user, cmd, cwd=None, env=None)
                if res["rc"] != 0:
                    raise RuntimeError(f"mkdir failed: {res['stderr'].strip()}")
                return self.make_response(item, {"mkdir": True, "path": remote_path})
            else:
                res = self._sftp_mkdir(host, port, user, password, remote_path, exist_ok=exist_ok)
                return self.make_response(item, res)
        elif svc == "ftp" or int(port) in [21, 990]:
            res = self._ftp_mkdir(host, port, user, password, remote_path, exist_ok=exist_ok)
            return self.make_response(item, res)
        else:
            return self.make_response(item, f"srv_mkdir not supported for port {port} ({svc}).")

    def cmd_srv_stat(self, item: dict) -> dict:
        p = item.get("params", {}) or {}
        server = p.get("server")
        port = int(p.get("port") or 0)
        remote_path = p.get("remote_path")
        if not (server and port and remote_path):
            return self.make_response(item, "Params 'server', 'port' and 'remote_path' required.")
        cfg = self._server_config(server, port)
        host = cfg.get("server")
        user = cfg.get("login")
        password = cfg.get("password")
        svc = self._service_from_port(port)
        if svc in ["ssh", "sftp", "unknown"] or int(port) == 22:
            if self._prefer_system_ssh():
                # Try 'stat' then fallback to test
                cmd = f"(stat -c '%F|%s|%Y' {self._shell_quote(remote_path)} 2>/dev/null) || echo 'UNKNOWN|||';"
                res = self._ssh_exec_system(host, port, user, cmd, cwd=None, env=None)
                if res["rc"] != 0:
                    raise RuntimeError(f"stat failed: {res['stderr'].strip()}")
                line = res["stdout"].strip().splitlines()[-1] if res["stdout"] else "UNKNOWN|||"
                kind, size, mtime = (line.split("|") + ["", "", ""])[:3]
                ftype = "file"
                if "directory" in kind.lower():
                    ftype = "dir"
                elif "link" in kind.lower():
                    ftype = "link"
                size_val = None
                try:
                    size_val = int(size) if size else None
                except Exception:
                    pass
                mtime_val = None
                try:
                    mtime_val = int(mtime) if mtime else None
                except Exception:
                    pass
                return self.make_response(item,
                                          {"path": remote_path, "type": ftype, "size": size_val, "mtime": mtime_val})
            else:
                res = self._sftp_stat(host, port, user, password, remote_path)
                return self.make_response(item, res)
        elif svc == "ftp" or int(port) in [21, 990]:
            res = self._ftp_stat(host, port, user, password, remote_path)
            return self.make_response(item, res)
        else:
            return self.make_response(item, f"srv_stat not supported for port {port} ({svc}).")

    def cmd_smtp_send(self, item: dict) -> dict:
        p = item.get("params", {}) or {}
        server = p.get("server")
        port = int(p.get("port") or 0)
        if not (server and port):
            return self.make_response(item, "Params 'server' and 'port' required.")
        cfg = self._server_config(server, port)
        host = cfg.get("server")
        user = cfg.get("login")
        password = cfg.get("password")

        mail = {
            "from_addr": p.get("from_addr"),
            "to": p.get("to"),
            "cc": p.get("cc"),
            "bcc": p.get("bcc"),
            "subject": p.get("subject"),
            "body": p.get("body"),
            "html": bool(p.get("html") or False),
            "attachments": p.get("attachments") or [],
            "use_tls": p.get("use_tls"),
            "use_ssl": p.get("use_ssl"),
        }
        res = self._smtp_send_mail(host, port, user, password, mail)
        return self.make_response(item, res)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2026.01.03 00:00:00                  #
# ================================================== #

from PySide6.QtCore import Qt
from PySide6.QtGui import QAction, QIcon, QKeySequence, QFontMetrics
from PySide6.QtWidgets import QTextEdit

from pygpt_net.core.text.finder import Finder
from pygpt_net.utils import trans


class BaseCodeEditor(QTextEdit):

    _ICON_VOLUME = QIcon(":/icons/volume.svg")
    _ICON_SAVE = QIcon(":/icons/save.svg")
    _ICON_SEARCH = QIcon(":/icons/search.svg")
    _ICON_CLOSE = QIcon(":/icons/close.svg")
    _FIND_SEQ = QKeySequence("Ctrl+F")

    def __init__(self, window=None):
        """
        Base code editor

        :param window: main window
        """
        super().__init__(window)
        self.window = window
        self.finder = Finder(window, self)
        self.setReadOnly(True)
        self.setAcceptRichText(False)
        self.value = 12
        self.max_font_size = 42
        self.min_font_size = 8
        self.setProperty('class', 'code-editor')
        self.default_stylesheet = ""
        self.setStyleSheet(self.default_stylesheet)
        self.excluded_copy_to = []
        self.textChanged.connect(self.text_changed)

        metrics = QFontMetrics(self.font())
        space_width = metrics.horizontalAdvance(" ")
        self.setTabStopDistance(4 * space_width)

    def text_changed(self):
        self.finder.text_changed()

    def update_stylesheet(self, data: str):
        self.setStyleSheet(self.default_stylesheet + data)

    def contextMenuEvent(self, event):
        menu = self.createStandardContextMenu()
        cursor = self.textCursor()
        selected_text = cursor.selectedText()

        if selected_text:
            plain_text = cursor.selection().toPlainText()

            action = QAction(self._ICON_VOLUME, trans('text.context_menu.audio.read'), menu)
            action.triggered.connect(self.audio_read_selection)
            menu.addAction(action)

            copy_to_menu = self.window.ui.context_menu.get_copy_to_menu(
                menu,
                selected_text,
                excluded=self.excluded_copy_to
            )
            try:
                copy_to_menu.setParent(menu)
            except Exception:
                pass
            menu.addMenu(copy_to_menu)

            action = QAction(self._ICON_SAVE, trans('action.save_selection_as'), menu)
            action.triggered.connect(
                lambda: self.window.controller.chat.common.save_text(plain_text))
            menu.addAction(action)
        else:
            action = QAction(self._ICON_SAVE, trans('action.save_as'), menu)
            action.triggered.connect(
                lambda: self.window.controller.chat.common.save_text(self.toPlainText()))
            menu.addAction(action)

        # Add zoom submenu
        zoom_menu = self.window.ui.context_menu.get_zoom_menu(self, "editor", self.value, self.on_zoom_changed)
        menu.addMenu(zoom_menu)

        action = QAction(self._ICON_SEARCH, trans('text.context_menu.find'), menu)
        action.triggered.connect(self.find_open)
        action.setShortcut(self._FIND_SEQ)
        menu.addAction(action)

        action = QAction(self._ICON_CLOSE, trans('action.clear'), menu)
        action.triggered.connect(self.clear_content)
        menu.addAction(action)

        menu.exec(event.globalPos())
        menu.deleteLater()

    def clear_content(self):
        self.clear()

    def audio_read_selection(self):
        self.window.controller.audio.read_text(self.textCursor().selectedText())

    def find_open(self):
        self.window.controller.finder.open(self.finder)

    def on_update(self):
        self.finder.clear()

    def on_destroy(self):
        self.window.controller.finder.unset(self.finder)

    def keyPressEvent(self, e):
        if e.key() == Qt.Key_F and e.modifiers() & Qt.ControlModifier:
            self.find_open()
        else:
            super().keyPressEvent(e)

    def wheelEvent(self, event):
        if event.modifiers() & Qt.ControlModifier:
            prev = self.value
            if event.angleDelta().y() > 0:
                if self.value < self.max_font_size:
                    self.value += 1
            else:
                if self.value > self.min_font_size:
                    self.value -= 1

            if self.value != prev:
                self.update_stylesheet(f"QTextEdit {{ font-size: {self.value}px }};")
            event.accept()
        else:
            super().wheelEvent(event)

    def on_zoom_changed(self, value: int):
        """
        On font size changed

        :param value: New font size
        """
        self.value = value
        self.update_stylesheet(f"QTextEdit {{ font-size: {value}px }};")

    def focusInEvent(self, e):
        super().focusInEvent(e)
        self.window.controller.finder.focus_in(self.finder)


class CodeEditor(BaseCodeEditor):
    def __init__(self, window=None):
        """
        Code editor

        :param window: main window
        """
        super().__init__(window)